from typing import List

from ..types import ColumnNode
from ..types.graph import Node, Edge, Graph


def create_dependency_graph(columns: list[ColumnNode]) -> Graph:
    """
    Convert column definitions into a graph structure with nodes and edges.
    """
    nodes: List[Node] = [
        Node(
            id=col.id,
            name=col.name,
            type=col.data_type,
            sensitivity=col.sensitivity,
            schema=col.schema,
            stage=col.stage,
            node_category="column",
        )
        for col in columns
    ]

    edges: List[Edge] = [
        Edge(source=source, target=col.id)
        for col in columns
        for source in col.parent_ids
    ]

    return Graph(nodes=nodes, edges=edges)



from enum import StrEnum, auto
import re
from typing import Dict, Literal

import polars as pl

from ..ref import parse_column_name

PREFIX = ""

"""
---------------------------------------------------------------------------------------------
    Cleaning functions
---------------------------------------------------------------------------------------------
These functions are used to clean the data in the columns - usually by converting to a 
correct data type
"""


class SanitiseLevel(StrEnum):
    FULL = auto()
    LOWERCASE = auto()
    TRIM = auto()


# Unfortunately, we can't use the enum as a type hint for the sanitise attribute
type SanitiseLevelType = Literal["full", "lowercase", "trim"] | None


def sanitise_string_col(
    col: str, *, sanitize_level: SanitiseLevelType = None, join_char: str = "_"
) -> pl.Expr:
    if sanitize_level == SanitiseLevel.FULL:
        return full_sanitise_string_col(col, join_char=join_char)
    elif sanitize_level == SanitiseLevel.LOWERCASE:
        return lower_sanitise_string_col(col)
    elif sanitize_level == SanitiseLevel.TRIM:
        return trim_sanitise_string_col(col)
    else:
        return pl.col(col)


def _sanitisation_fix(exp: pl.Expr, incorrect: str, corrected: str):
    # Can't do look behind regex in polars, so have to declare each option separately
    return (
        exp.str.replace_all(rf"^{incorrect}$", f"{corrected}", literal=False)
        .str.replace_all(rf"^{incorrect}_", f"{corrected}_", literal=False)
        .str.replace_all(rf"_{incorrect}$", f"_{corrected}", literal=False)
        .str.replace_all(rf"_{incorrect}_", f"_{corrected}_", literal=False)
    )


COMMON_MISTAKES = (
    ("i_d", "id"),
    ("u_r_l", "url"),
    ("h_t_t_p", "http"),
    ("h_t_m_l", "html"),
    ("c_s_v", "csv"),
    ("x_l_s_x", "xlsx"),
    ("p_d_f", "pdf"),
    ("a_p_i", "api"),
    ("d_b", "db"),
    ("u_i", "ui"),
    ("u_x", "ux"),
    ("s_q_l", "sql"),
    ("j_s_o_n", "json"),
    ("x_m_l", "xml"),
    ("s_s_l", "ssl"),
    ("t_l_s", "tls"),
    ("n_z", "nz"),
    ("u_k", "uk"),
    ("u_s", "us"),
    ("e_u", "eu"),
    ("m_ori", "maori"),
    ("wh_nau", "whanau"),
    ("p_keh_", "pakeha"),
    ("t_ne", "tane"),
    ("w_hine", "wahine"),
    ("hap_", "hapu"),
    ("w_nanga", "wananga"),
    ("k_inga", "kainga"),
    ("k_rero", "korero"),
)


def trim_sanitise_string_col(col: str) -> pl.Expr:
    """
    Remove leading and trailing spaces
    """
    return pl.col(col).cast(pl.String).str.strip_chars()


def lower_sanitise_string_col(col: str) -> pl.Expr:
    """
    Convert to lowercase
    """
    return trim_sanitise_string_col(col).str.to_lowercase()


def full_sanitise_string_col(
    col: str,
    join_char="_",
) -> pl.Expr:
    new_col = (
        lower_sanitise_string_col(col)
        # Replace special characters with their plain letter counterparts
        .str.replace_all("ā", "a", literal=True)
        .str.replace_all("ē", "e", literal=True)
        .str.replace_all("ī", "i", literal=True)
        .str.replace_all("ō", "o", literal=True)
        .str.replace_all("ū", "u", literal=True)
        # Replace spaces, question marks, slashes, and line breaks with underscores
        .str.replace_all(r"[ ?/\n\r]", "_", literal=False)
        # Remove any non-alphanumeric characters (except underscores)
        .str.replace_all(r"[^a-z0-9_]", "_", literal=False)
    )
    # handle common mis-coded words
    for incorrect, corrected in COMMON_MISTAKES:
        new_col = _sanitisation_fix(new_col, incorrect, corrected)

    new_col = (
        new_col
        # Replace multiple underscores with a single one
        .str.replace_all(r"_+", "_", literal=False)
        # Strip leading and trailing spaces (technically underscores...)
        .str.strip_chars("_")
        # Change underscores back to spaces
        .str.replace_all("_", join_char, literal=True)
    )
    return new_col


def sanitise_scalar_string(value: str, join_char: str = "_") -> str:
    """
    Sanitise a string value by converting to lowercase, replacing special characters with their plain letter counterparts,
    and replacing spaces, question marks, slashes, and line breaks with underscores.
    Also handles common mis-coded words.
    """
    if not isinstance(value, str):
        value = str(value)

    # Replace special characters with their plain letter counterparts
    value = (
        value.strip()
        .lower()
        .replace("ā", "a")
        .replace("ē", "e")
        .replace("ī", "i")
        .replace("ō", "o")
        .replace("ū", "u")
    )
    # Replace spaces, question marks, slashes, and line breaks with underscores
    value = re.sub(r"[ ?/\n\r]", "_", value)
    # Replace any non-alphanumeric characters (except underscores)
    value = re.sub(r"[^a-z0-9_]", "_", value)
    # Handle common mis-sanitised words
    value = fix_sanitisation_mistakes(value)
    # Replace multiple underscores with a single one
    value = re.sub(r"_+", "_", value)
    # Strip leading and trailing underscores/spaces
    value = value.strip("_ ")

    # Finally, replace underscores with the specified join character
    return value.replace("_", join_char)


def clean_enum_col(
    col: str,
    enum_def=Dict[str, str],
    *,
    prefix: str = PREFIX,
    alias: str = None,
    default="No Data",
    sanitize: bool = True,
) -> pl.Expr:
    unique_values = set(enum_def.values())
    unique_values.add(default)
    new_col = pl.col(col) if not sanitize else full_sanitise_string_col(col)

    return new_col.replace(
        enum_def,
        default=default,
        return_dtype=pl.Enum(list(unique_values)),
    ).alias(alias if alias else f"{prefix}{col}")


def clean_boolean_col(
    col: str,
    *,
    prefix: str = PREFIX,
    alias: str = None,
    true_values=("yes", "true", "1"),
    false_values=("no", "false", "0"),
    sanitize: bool = True,
) -> pl.Expr:
    new_col = pl.col(col) if not sanitize else full_sanitise_string_col(col)

    return new_col.replace(
        {**{val: True for val in true_values}, **{val: False for val in false_values}},
        default=None,
        return_dtype=pl.Boolean,
    ).alias(alias if alias else f"{prefix}{col}")


def fix_sanitisation_mistakes(text, mistakes=COMMON_MISTAKES):
    # Handle cases where the incorrect term appears:
    # 1. As a complete word by itself
    # 2. At the start of a compound word
    # 3. At the end of a compound word
    # 4. In the middle of a compound word
    for incorrect, corrected in mistakes:
        pattern = rf"(^|_)({re.escape(incorrect)})($|_)"
        replacement = rf"\1{corrected}\3"
        text = re.sub(pattern, replacement, text)

    # value = text
    # for incorrect, corrected in COMMON_MISTAKES:
    #     value = re.sub(rf"^{incorrect}$", corrected, value)
    #     value = re.sub(rf"^{incorrect}_", f"{corrected}_", value)
    #     value = re.sub(rf"_{incorrect}$", f"_{corrected}", value)
    #     value = re.sub(rf"_{incorrect}_", f"_{corrected}_", value)
    return text


def sanitise_column_name(column: str, join_char="_"):
    """
    --------------------------------------------------------------------------------------------------
      Sanitise column name
    --------------------------------------------------------------------------------------------------
    This function takes a string (assumed to be a column name) and sanitises it by:

    - Removing the prefix/suffix (if any)
    - Converting camel case words to snake case
    - Lowercasing the string
    - Replacing special characters with their plain letter counterparts
    - Replacing common incorrect words with their correct versions
    - Replacing spaces, dashes, question marks, slashes, and line breaks with underscores
    - Removing any non-alphanumeric characters (except underscores)
    - Replacing multiple underscores with a single one
    - Stripping leading and trailing underscores/spaces
    - Appending the prefix/suffix back to the string (if any)

    :param column: The original string/column name
    :param join_char: The character to use to join words in the column headers (default: "_")
    :return: A sanitised string/column name
    """
    ref = parse_column_name(column)
    # Convert camel case to snake case, then lowercase
    col = sanitise_scalar_string(
        re.sub(r"(?<!^)(?=[A-Z])", "_", ref.base_name).lower(),
        join_char=join_char
    )

    return ref.with_base_name(col).name


def sanitise_column_headers(lf: pl.LazyFrame, join_char="_") -> pl.LazyFrame:
    """
    --------------------------------------------------------------------------------------------------
      Sanitise column headers
    --------------------------------------------------------------------------------------------------
    This function takes a LazyFrame and sanitises the column headers by:

    - Lowercasing the column name
    - Replacing special characters with their plain letter counterparts
    - Replacing common incorrect words with their correct versions
    - Replacing spaces, dashes, question marks, slashes, and line breaks with underscores
    - Removing any non-alphanumeric characters (except underscores)
    - Replacing multiple underscores with a single one
    - Stripping leading and trailing underscores/spaces

    :param lf: The Polars LazyFrame to sanitise
    :param join_char: The character to use to join words in the column headers (default: "_")
    :return: A Polars LazyFrame with the column headers sanitised
    """

    return lf.rename(
        # {col: sanitise(col) for col in lf.collect_schema().names() if not re.match(_INTERNAL_COL_PATTERN, col)})
        {
            col: sanitise_column_name(col, join_char)
            for col in lf.collect_schema().names()
        }
    )



from typing import TYPE_CHECKING

from ...types import ColumnNode
from ...types.sensitivity import Sensitivity
from ...types.validation.threshold import Threshold, ThresholdLevel, CheckLevel
from ...types.validation.validation_rule import ColumnValidationRule

if TYPE_CHECKING:
    from ...ref import ColumnRef
    from .check_base import RowCheckBase


def create_validation_rule(
    check_method_id: str,
    message: str,
    *,
    source_column: str,
    alias: str,
    schema: str,
    thresholds: Threshold | None,
) -> ColumnValidationRule | None:
    """Create a validation rule for a check"""
    if not thresholds:
        return None

    return ColumnValidationRule(
        check_id=check_method_id,
        message=message,
        thresholds=thresholds,
        source_column=source_column,
        check_column=alias,
        schema=schema,
    )


def _get_threshold_level(thresholds: Threshold | None) -> CheckLevel:
    # We default to WARNING if no thresholds are provided (otherwise you can't differentiate between a pass and a fail)
    if thresholds is None:
        return ThresholdLevel.WARNING.level

    res = thresholds.resolve(1, 1)

    if res.level == ThresholdLevel.PASS.level:
        return ThresholdLevel.WARNING.level

    return res.level


def get_validation_node(
    check: "RowCheckBase",
    *,
    source_column: "ColumnRef",
    thresholds: Threshold | None,
    check_on_cleaned: bool = False,
    alias: str | None = None,
    label: str | None = None,
) -> ColumnNode:
    """Create a validation node for a check"""
    label = label if label else check.method_id()
    _ref = source_column.clean() if check_on_cleaned else source_column

    alias = alias if alias else _ref.check(label).name

    assert alias != source_column.name, (
        f"Check column name must be different from column - {source_column.name}"
    )
    assert alias != _ref.name, (
        f"Check column name must be different from cleaned column - {_ref.name}"
    )

    expr = check._check_expression(
        _ref.name, alias, level=_get_threshold_level(thresholds)
    )
    validation_rule = create_validation_rule(
        check_method_id=check.method_id(),
        message=check.message,
        source_column=source_column.name,
        alias=alias,
        schema=source_column.schema,
        thresholds=thresholds,
    )

    return ColumnNode(
        id=_ref.check(label).id,
        name=alias,
        data_type="boolean",
        column_type="validation",
        schema=source_column.schema,
        stage=source_column.stage,
        sensitivity=Sensitivity.NON_SENSITIVE,
        expression=expr,
        validation_rule=validation_rule,
    )



import pytest
import polars as pl
from zeolite.types import ColumnNode, Sensitivity
from zeolite.types.validation.threshold import Threshold, ThresholdLevel


@pytest.fixture
def sample_df():
    """Sample DataFrame for testing"""
    return pl.DataFrame(
        {
            "string_col": ["value1", None, "", "value2"],
            "number_col": [1, 2, None, 4],
            "boolean_col": ["yes", "no", "true", "invalid"],
        }
    )


@pytest.fixture
def sample_threshold():
    """Sample threshold configuration"""
    return Threshold(
        warning=0.1,  # 10% threshold
        error=0.5,  # 50% threshold
        reject=0.8,  # 80% threshold
    )


@pytest.fixture
def sample_column_node():
    """Sample column node for testing"""
    return ColumnNode(
        id="test::sample",
        name="sample",
        data_type="string",
        column_type="source",
        sensitivity=Sensitivity.NON_SENSITIVE,
        schema="test",
        stage=None,
        expression=None,
        validation_rule=None,
    )


@pytest.fixture
def sample_error_levels():
    """Sample error levels for testing"""
    return {
        "debug": ThresholdLevel.DEBUG.level,
        "warning": ThresholdLevel.WARNING.level,
        "error": ThresholdLevel.ERROR.level,
        "reject": ThresholdLevel.REJECT.level,
    }



from typing import Literal

type ColumnDataType = Literal[
    "string",
    "id",
    "number",
    "integer",
    "float",
    "decimal",
    "boolean",
    "date",
    "datetime",
    "enum",
    "sanitised_string",
    "unknown",
]



import polars as pl


def calc_financial_year_dates(date_col: str, *, prefix="") -> list[pl.Expr]:
    return [
        pl.col(date_col).dt.month().alias(f"{prefix}month"),
        pl.col(date_col).dt.to_string("%B").alias(f"{prefix}month_label"),
        pl.col(date_col).dt.year().alias(f"{prefix}calendar_year"),
        # ----------------------------------------------
        pl.when(pl.col(date_col).dt.quarter() > 2)
        .then(pl.col(date_col).dt.quarter() - 2)
        .otherwise(pl.col(date_col).dt.quarter() + 2)
        .alias(f"{prefix}financial_quarter"),
        # ----------------------------------------------
        pl.concat_str(
            [
                pl.col(date_col).dt.offset_by("-6mo").dt.year(),
                pl.col(date_col).dt.offset_by("6mo").dt.year(),
            ],
            separator="/",
        ).alias(f"{prefix}financial_year"),
        # ----------------------------------------------
        pl.col(date_col)
        .dt.month_start()
        .cast(pl.Date)
        .alias(f"{prefix}month_start_date"),
        pl.col(date_col).dt.month_end().cast(pl.Date).alias(f"{prefix}month_end_date"),
        pl.date(pl.col(date_col).dt.offset_by("-6mo").dt.year(), 7, 1).alias(
            f"{prefix}financial_year_start_date"
        ),
        pl.date(pl.col(date_col).dt.offset_by("6mo").dt.year(), 6, 30).alias(
            f"{prefix}financial_year_end_date"
        ),
        # fy_start=pl.when(pl.col("a").dt.quarter() > 2).then(pl.date(pl.col("a").dt.year(), 7, 1)).otherwise(
    ]



import json
from datetime import datetime
from enum import StrEnum
from typing import Any, List
import polars as pl

from ..types import ColumnNode
from ..types.graph import Graph
from .lineage import create_dependency_graph
from .optimise_calc_stages import generate_optimized_stages
from .verify_registry import verify_column_registry_integrity


# %%
class ColumnRegistry:
    def __init__(self, columns: list[ColumnNode] = None):
        self.by_id: dict[str, ColumnNode] = {}
        self.by_name: dict[str, ColumnNode] = {}

        cols = columns if columns is not None else []
        self.extend(cols)

    # ---------------------------------------------------------------------------------
    # Registry setup/management
    # ---------------------------------------------------------------------------------
    def append(self, column: ColumnNode) -> None:
        self._append(column)
        self._post_update()

    def extend(self, columns: list[ColumnNode]) -> None:
        for c in columns:
            self._append(c)
        self._post_update()

    def get_by_id(self, id: str) -> ColumnNode | None:
        return self.by_id.get(id)

    def get_by_name(self, name: str) -> ColumnNode | None:
        return self.by_name.get(name)

    def remove(self, column: ColumnNode) -> None:
        self.by_id.pop(column.id, None)
        self.by_name.pop(column.name, None)

    def get_all_ids(self):
        return self.by_id.keys()

    def _append(self, column: ColumnNode) -> None:
        assert column.id not in self.by_id, f"Duplicate column ID: {column.id}"
        assert column.name not in self.by_name, f"Duplicate column name: {column.name}"
        self.by_id[column.id] = column
        self.by_name[column.name] = column

    def _replace(self, column: ColumnNode) -> None:
        self.by_id[column.id] = column
        self.by_name[column.name] = column

    def _post_update(self) -> None:
        """
        When columns have been added to the registry, this method
        maps the parent_column names to the actual column/node IDs.
        """

        for col in self.nodes():
            if len(col.parent_columns) > 0:
                # self._replace(col.with_parent_ids({self.by_name[source].id if source in self.by_name else source for source in col.parent_columns}))

                # Only include source IDs for columns that exist in the registry
                matched_parent_ids = {
                    self.by_name[source].id
                    for source in col.parent_columns
                    if source in self.by_name
                }
                self._replace(col.with_parent_ids(matched_parent_ids))

    # ---------------------------------------------------------------------------------
    # Special methods
    # ---------------------------------------------------------------------------------
    def lineage(self) -> Graph:
        dependency_graph = create_dependency_graph(self.nodes())

        # return json.dumps(dependency_graph, cls=_SchemaJsonEncoder)
        return dependency_graph

    def verify_integrity(self) -> bool:
        return verify_column_registry_integrity(self.nodes())

    def get_execution_stages(self) -> List[List[ColumnNode]]:
        return generate_optimized_stages(self.nodes())

    # ----------------------------------------------------------------------------------
    #  Iterator methods
    # ----------------------------------------------------------------------------------
    def nodes(self) -> list[ColumnNode]:
        return list(self.by_id.values())

    def __iter__(self):
        return iter(self.nodes())


class _SchemaJsonEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, StrEnum):
            return obj.value
        if isinstance(obj, set):
            return list(obj)
        if isinstance(obj, datetime):
            return obj.isoformat()
        if isinstance(obj, pl.Expr):
            return str(obj)
        if isinstance(obj, ColumnRegistry):
            return obj.by_name

        # if isinstance(obj, Graph):
        #     return asdict(obj)

        # if isinstance(obj, DerivedColumns):
        #     cols = [c.meta.output_name() for c in obj.args]
        #     cols.extend([k for k in obj.kwargs.keys()])
        #     return cols
        return super().default(obj)



from dataclasses import dataclass, field, replace
from datetime import datetime
from typing import Literal

from .validation.threshold import CheckLevel


@dataclass(frozen=True)
class SchemaValidationError:
    schema: str

    error: str
    level: CheckLevel
    message: str
    column: str | None = None
    fraction_failed: str | None = None
    count_failed: str | None = None
    category: Literal["file", "schema", "logic"] = field(init=False, default="unknown")

    # Meta fields
    source: str | None = None
    batch: datetime | None = None
    period: str | None = None
    year: str | None = None

    def with_meta(
        self,
        *,
        source: str | None = None,
        period: str | None = None,
        year: str | None = None,
        batch: datetime | None = None,
    ) -> "SchemaValidationError":
        return replace(
            self,
            **{
                "source": source or self.source,
                "batch": batch or self.batch,
                "period": period or self.period,
                "year": year or self.year,
            },
        )


@dataclass(frozen=True)
class UnknownValidationError(SchemaValidationError):
    category = "unknown"
    level: CheckLevel
    column: None = field(init=False, default=None)


@dataclass(frozen=True)
class FileValidationError(SchemaValidationError):
    category = "file"
    level: CheckLevel
    column: None = field(init=False, default=None)


@dataclass(frozen=True)
class StructureValidationError(SchemaValidationError):
    category = "schema_structure"
    level: CheckLevel
    column: str | None = None


@dataclass(frozen=True)
class DataValidationError(SchemaValidationError):
    category = "row_data"
    level: CheckLevel
    column: str



from .schema import schema
from .column import (
    col,
    str_col,
    bool_col,
    date_col,
    int_col,
    float_col,
    derived_col,
    derived_custom_check,
    meta_col,
)
from zeolite.ref import (
    ref,
    ref_meta,
    ref_derived,
    ref_custom_check,
)
from .column.validation import (
    IsValueEmpty,
    IsValueDuplicated,
    IsValueInvalidDate,
    IsValueEqualTo,
)

from .types.sensitivity import Sensitivity
from .types.validation.threshold import Threshold

__all__ = [
    "schema",
    "col",
    "str_col",
    "bool_col",
    "date_col",
    "int_col",
    "float_col",
    "derived_col",
    "derived_custom_check",
    "meta_col",
    "check_is_value_empty",
    "check_is_value_duplicated",
    "check_is_value_invalid_date",
    "check_is_value_equal_to",
    "ref",
    "ref_meta",
    "ref_derived",
    "ref_custom_check",
    "Sensitivity",
    "Threshold",
]


check_is_value_empty = IsValueEmpty
"""
Validation: Check if a column value is empty/null.
"""

check_is_value_duplicated = IsValueDuplicated
"""
Validation: Check if a column value is duplicated.
"""

check_is_value_invalid_date = IsValueInvalidDate
"""
Validation: Check if a column value is an invalid date.
"""

check_is_value_equal_to = IsValueEqualTo
"""
Validation: Check if a column value is equal to a specified value.
"""




"""
Basic examples of using the Zeolite data validation framework.
This file demonstrates how to create schemas and define column validations.
"""

import zeolite as z

# ---------------------------------------------------------------------------------
# Example 1: Basic Schema Definition
# ---------------------------------------------------------------------------------
simple_schema = z.schema("basic").columns(
    z.col("id").validations(z.check_is_value_empty(), z.check_is_value_duplicated()),
    z.col("age").data_type("integer"),
    z.str_col("name").aliases("full_name", "person_name"),
    z.date_col("birthdate"),
)

# ---------------------------------------------------------------------------------
# Example 2: Specialized Column Types with Validations
# ---------------------------------------------------------------------------------
detailed_schema = z.schema("detailed").columns(
    # ID column with validations
    z.col("id")
    .data_type("id")
    .required()
    .validations(z.check_is_value_empty(), z.check_is_value_duplicated()),
    # String column
    z.str_col("name")
    .sensitivity(z.Sensitivity.IDENTIFIABLE)
    .validations(z.check_is_value_empty()),
    # Integer column
    z.int_col("age").validations(z.check_is_value_empty()),
    # Date column
    z.date_col("enrollment_date")
    .aliases("date_enrolled", "enrollment")
    .validations(z.check_is_value_invalid_date(check_on_cleaned=False)),
    # Boolean column
    z.bool_col("is_active"),
    # Float column
    z.float_col("balance"),
)

# ---------------------------------------------------------------------------------
# Example 3: Column parsing/cleaning
# ---------------------------------------------------------------------------------

cleaned_schema = z.schema("cleaned_individual").columns(
    z.str_col("id")
    .clean(to="id", prefix="ORG_X::")
    .validations(
        z.check_is_value_empty(warning="any", error=0.1, reject=0.01),
        z.check_is_value_duplicated(check_on_cleaned=True, reject="any"),
    ),
    z.str_col("birthdate")
    .clean(to="date")
    .validations(z.check_is_value_invalid_date()),
    z.str_col("gender").clean(
        to="enum",
        enum_map={"m": "Male", "f": "Female", "female": "Female", "male": "Male"},
    ),
    z.str_col("is_active").clean(
        to="boolean", true_values={"yes", "active"}, false_values={"no", "inactive"}
    ),
    z.str_col("ethnicity").clean(to="sanitised_string"),
    z.str_col("ethnicity_2").clean(to="sanitised_string"),
    z.derived_col(
        "is_maori",
        function=(
            z.ref("ethnicity").clean().col.eq("maori")
            | z.ref("ethnicity_2").clean().col.eq("maori")
        ),
        data_type="boolean",
    ),
)

# ---------------------------------------------------------------------------------
# Example 4: Custom Validations
# ---------------------------------------------------------------------------------
schema_with_custom_check = z.schema("custom_checks").columns(
    z.int_col("score"),
    # Custom validation: ensure score is between 0 and 100
    z.derived_custom_check(
        name="valid_score_range",
        function=(z.ref("score").col.lt(0) & z.ref("score").col.gt(100)),
        message="Score must be between 0 and 100",
        thresholds=z.Threshold(reject=0.01),  # Reject if more than 1% fail
    ),
)



# %%
from dataclasses import dataclass, field
from typing import Literal
from enum import Enum

# %%
type CheckThreshold = bool | int | float | Literal["any", "all"] | None

type CheckLevel = Literal["warning", "error", "reject", "pass", "debug"]


class ThresholdLevel(Enum):
    REJECT = ("reject", 6)
    # CRITICAL = ("critical", 5)
    ERROR = ("error", 4)
    WARNING = ("warning", 3)
    # INFO = ("info", 2)
    DEBUG = ("debug", 1)
    PASS = ("pass", 0)

    def __init__(self, level: CheckLevel, priority: int):
        self.level = level
        self.priority = priority


@dataclass(frozen=True)
class ThresholdResult:
    level: CheckLevel
    fraction_failed: float | None = None
    count_failed: int | None = None


def _parse_threshold_value(
    value: CheckThreshold, level_name: str
) -> tuple[float | None, int | None]:
    """Parse a threshold value into fraction and count components."""
    if value == 1 and value is not True:
        raise ValueError(
            f"`{level_name}` cannot be 1 as it is ambiguous, use `'all'` or `'any'` instead."
        )

    if value == "any" or value is True or value == 0:
        return 0, 1
    elif value == "all":
        return 1, None
    elif isinstance(value, (int, float)) and 0 < value < 1:
        return value, None
    elif isinstance(value, (int, float)) and value > 1:
        return None, round(value)
    elif isinstance(value, (int, float)) and value < 0:
        raise ValueError(f"Negative values are not allowed for `{level_name}`.")
    elif isinstance(value, str):
        if value not in ["any", "all"]:
            raise ValueError(
                f"`{level_name}` must be `'any'` or `'all'` for string values"
            )
        return (0, 1) if value == "any" else (1, None)
    return None, None


@dataclass()
class Threshold:
    """
    Represents a set of thresholds for different levels of checks and evaluates
    the result based on failed and total rows.

    The Threshold class is used to define thresholds for various levels of checks,
    such as debug, warning, error, and reject. Each threshold can define criteria
    based on either the fraction of failed rows or the absolute count of failed rows.
    Using these thresholds, the class can determine the appropriate threshold level
    for a given dataset based on the number of failed rows and total rows.

    Attributes:
        debug (CheckThreshold): The threshold config for the debug level.
        warning (CheckThreshold): The threshold config for the warning level.
        error (CheckThreshold): The threshold config for the error level.
        reject (CheckThreshold): The threshold config for the reject level.

    """
    debug: CheckThreshold = None
    warning: CheckThreshold = None
    error: CheckThreshold = None
    reject: CheckThreshold = None

    _thresholds: dict[ThresholdLevel, tuple[float | None, int | None]] = field(
        init=False
    )

    def __post_init__(self):
        if (
            (self.debug is None)
            and (self.warning is None)
            and (self.error is None)
            and (self.reject is None)
        ):
            raise ValueError(
                "At least one of `debug`, `warning`, `error`, or `reject` must be provided"
            )

        # Initialize thresholds dictionary with parsed values
        thresholds = {}
        for level in [
            ThresholdLevel.REJECT,
            ThresholdLevel.ERROR,
            ThresholdLevel.WARNING,
            ThresholdLevel.DEBUG,
        ]:
            value = getattr(self, level.level)
            if value is not None and value is not False:
                fraction, count = _parse_threshold_value(value, level.level)
                thresholds[level] = (fraction, count)

        object.__setattr__(self, "_thresholds", thresholds)

    def resolve(self, failed_rows: int, total_rows: int) -> ThresholdResult:
        """Resolve the threshold level based on failed rows and total rows."""
        if total_rows == 0:
            return ThresholdResult(level="pass")

        fraction_failed = failed_rows / total_rows if total_rows > 0 else 1
        rows_failed = failed_rows if failed_rows > 0 else 0

        # Check each threshold level in priority order
        for level in sorted(ThresholdLevel, key=lambda x: x.priority, reverse=True):
            if level == ThresholdLevel.PASS:
                continue

            fraction, count = self._thresholds.get(level, (None, None))

            # Check if threshold is met
            if (fraction is not None and fraction_failed >= fraction) or (
                count is not None and rows_failed >= count
            ):
                return ThresholdResult(
                    level=level.level,
                    fraction_failed=fraction_failed,
                    count_failed=rows_failed,
                )

        return ThresholdResult(
            level="pass", fraction_failed=fraction_failed, count_failed=rows_failed
        )



from dataclasses import dataclass, field, replace
from typing import List, Literal
from collections.abc import Iterable

from polars import LazyFrame

from ._utils.normalise import normalise_column_headers
from ..types import (
    ColumnNode,
    SourceColDef,
    ThresholdLevel,
    ValidationResult,
    ProcessingFailure,
    ProcessingSuccess,
)
from .._utils.args import flatten_args
from .._utils.sanitize import sanitise_column_name
from ..column import ColumnSchema
from ..registry import ColumnRegistry, generate_optimized_stages


@dataclass(frozen=True, kw_only=True)
class _SchemaParams:
    name: str
    is_required: bool = False
    stage: str | None = None
    source_columns: dict[str, SourceColDef] = field(default_factory=dict)
    registry: ColumnRegistry = field(default_factory=ColumnRegistry)


class TableSchema:
    """
    Defines a table schema for data validation and processing.

    Parameters:
        name (str): Name of the schema.
        columns (List[ColumnSchema] | dict[str, ColumnSchema], optional): List of column schemas.
        is_required (bool): Whether the schema is required.
        stage (str, optional): Processing stage.

    Usage:
        demo_schema = z.schema("demo", columns=[...])
    """

    def __init__(
        self,
        name: str,
        columns: Iterable[ColumnSchema] | dict[str, ColumnSchema] | None = None,
        *,
        is_required: bool = False,
        stage: str | None = None,
    ):
        cols = _parse_columns_into_list(columns)
        nodes = _cols_to_nodes(schema=name, stage=stage, columns=cols)

        self._params = _SchemaParams(
            name=name,
            is_required=is_required,
            stage=stage,
            registry=ColumnRegistry(nodes),
            source_columns=_cols_to_sources(cols),
        )

    def columns(
        self,
        *args: ColumnSchema | Iterable[ColumnSchema] | dict[str, ColumnSchema],
        method: Literal["merge", "replace"] = "merge",
        **kwargs: ColumnSchema,
    ) -> "TableSchema":
        """Add or replace columns in the schema.

        Args:
            *args: ColumnSchema objects, lists of ColumnSchema, or dicts mapping names to ColumnSchema
            method: Either "merge" to add to existing columns or "replace" to replace all columns
            **kwargs: Named ColumnSchema objects where the key is the column name

        Returns:
            TableSchema: A new schema with the updated columns
        """
        cols = _parse_columns_into_list(*args, **kwargs)

        if method == "merge":
            new_nodes = _cols_to_nodes(
                self._params.name, self._params.stage, columns=cols
            )
            new_registry = _reset_registry(self._params.registry.nodes() + new_nodes)
            new_source_columns = {
                **self._params.source_columns,
                **_cols_to_sources(cols),
            }
            return self._replace(
                registry=new_registry, source_columns=new_source_columns
            )
        elif method == "replace":
            new_registry = _reset_registry(
                _cols_to_nodes(self._params.name, self._params.stage, columns=cols)
            )
            new_source_columns = _cols_to_sources(cols)
            return self._replace(
                registry=new_registry, source_columns=new_source_columns
            )
        else:
            raise ValueError(f"Invalid method: {method}")

    def required(self, is_required: bool = True) -> "TableSchema":
        """Set the schema as required"""
        return self._replace(is_required=is_required)

    @property
    def name(self) -> str:
        """Get the name of the table schema"""
        return self._params.name

    @property
    def is_required(self) -> bool:
        """Get the required status of the table schema"""
        return self._params.is_required

    @property
    def stage(self) -> str:
        """Get the stage of the table schema"""
        return self._params.stage

    def step_1_normalise(
        self, lf: LazyFrame, *, source_name: str | None = None
    ) -> ValidationResult:
        """
        Normalises the provided Polars LazyFrame to ensure column headers align with the
        defined schema/column definitions. It will attempt to rename columns based on
        the column aliases, if column headers are missing it will add them, and drop any
        additional column not defined in the schema.

        Args:
            lf (LazyFrame): The input lazy frame to be normalised.
            source_name (str | None): Optional name of the source for populating error messages.

        Returns:
            ValidationResult: The result of the normalization process, including the
            normalized lazy frame and any errors encountered during the process.

        """
        source_columns = list(self._params.source_columns.values())
        return normalise_column_headers(
            lf.lazy(),
            schema_name=self._params.name,
            col_defs=source_columns,
            source_name=source_name,
        )

    def step_2_prepare(
        self, lf: LazyFrame, *, source_name: str | None = None
    ) -> ValidationResult:
        """
        Applies optimized transformations to the provided Polars LazyFrame (lf). This function
        uses the column definitions to create cleaned & derived/calculated columns, and create
        check columns based on the validation rules defined for each column.

        Args:
            lf (LazyFrame): The input LazyFrame to be processed.
            source_name (str | None): Optional name of the source for populating error messages.

        Returns:
            ValidationResult: The result of the preparation process, including the
            processed LazyFrame and any errors encountered during the process.
        """
        prepped = lf

        stages = generate_optimized_stages(self._params.registry.nodes())

        for stage in stages:
            prepped = prepped.with_columns([c.expression for c in stage])

        # TODO: Add a check to see if the processing stages succeed
        return ValidationResult(data=prepped, errors=[], reject=False)

    def step_3_validate(
        self, lf: LazyFrame, *, source_name: str | None = None
    ) -> ValidationResult:
        """
        Applies validation rules to the provided Polars LazyFrame (lf) based on the column definitions.
        This uses the check columns created in step 2 and applies the validation thresholds to determine
        the resulting validation status/level. The returned LazyFrame isn't modified, but should only be used
        if the entire validation process succeeds without any `reject` level errors.

        Args:
            lf (LazyFrame): The input LazyFrame to be validated.
            source_name (str | None): Optional name of the source for populating error messages.

        Returns:
            ValidationResult: The result of the validation process, including the
            validated LazyFrame and any errors encountered during the process.
        """
        validation_errors = []
        reject = False
        for c in self._params.registry.nodes():
            if c.validation_rule is None:
                continue
            check = c.validation_rule.validate(lf, source=source_name)
            if check is not None:
                validation_errors.append(check)
                if check.level == ThresholdLevel.REJECT.level:
                    reject = True

        return ValidationResult(data=lf, errors=validation_errors, reject=reject)

    def process_data(
        self, lf: LazyFrame, *, source_name: str | None = None
    ) -> ProcessingFailure | ProcessingSuccess:
        """
        Processes data through multiple stages: normalisation, preparation, and validation.
        Each stage processes the input data and checks for errors. If any stage fails,
        the process aborts, returning the stage results and a failure status. If all stages
        are successful, the method returns a success status with all intermediate and final
        results.

        Args:
            lf (LazyFrame): LazyFrame containing the data to be processed.
            source_name (str | None): Optional name of the source for populating error messages.

        Returns:
            ProcessingResult: If any stage (`normalise`, `prepare`, or `validate`) fails with a `reject`,
            returns a `ProcessingFailure` containing intermediate results, encountered errors, and
            the name of the failed stage. Otherwise, returns a `ProcessingSuccess` containing
            intermediate and final results, as well as collected errors during processing.
        """
        errors = []
        normalised = self.step_1_normalise(lf, source_name=source_name)
        errors.extend(normalised.errors)
        if normalised.reject:
            return ProcessingFailure(
                normalised=normalised.data,
                prepared=None,
                validated=None,
                errors=errors,
                failed_stage="normalise",
            )

        prepped = self.step_2_prepare(normalised.data, source_name=source_name)
        errors.extend(prepped.errors)
        if prepped.reject:
            return ProcessingFailure(
                normalised=normalised.data,
                prepared=prepped.data,
                validated=None,
                errors=errors,
                failed_stage="prepare",
            )

        valid = self.step_3_validate(prepped.data, source_name=source_name)
        errors.extend(valid.errors)

        if valid.reject:
            return ProcessingFailure(
                normalised=normalised.data,
                prepared=prepped.data,
                validated=None,
                errors=errors,
                failed_stage="validate",
            )

        # If we get here, the data is valid
        return ProcessingSuccess(
            normalised=normalised.data,
            prepared=prepped.data,
            validated=valid.data,
            errors=errors,
        )

    def _replace(self, **kwargs):
        if "registry" not in kwargs:
            stage = kwargs.get("stage", self._params.stage)
            name = kwargs.get("name", self._params.name)
            kwargs["registry"] = ColumnRegistry(
                [
                    replace(n, stage=stage, schema=name)
                    for n in self._params.registry.nodes()
                ]
            )

        new_params = replace(self._params, **kwargs)
        return TableSchema(name=new_params.name).__set_params(new_params)

    def __set_params(self, params: _SchemaParams) -> "TableSchema":
        self._params = params
        return self


def _cols_to_nodes(
    schema: str, stage: str | None = None, columns: List[ColumnSchema] | None = None
) -> List[ColumnNode]:
    nodes = []
    if columns is None:
        return nodes

    for c in columns:
        assert isinstance(c, ColumnSchema), (
            f"All columns must be a Column Schema definition - {c}"
        )
        assert c.ref.base_name is not None, """
        All Columns must have a name, either:
          - directly on creation, e.g. z.col("name")
          - assigning after creation, e.g. z.col().with_name("name")
          - defined through a keyword arg on schema.columns e.g. x.columns( name = z.col() )
          - defined through a dict key on schema.columns e.g. x.columns({ "name": z.col() })
        """
        nodes.extend(c.get_nodes(schema, stage))

    return nodes


def _cols_to_sources(columns: List[ColumnSchema] | None) -> dict[str, SourceColDef]:
    sources = {}
    aliases = {}
    if columns is None:
        return sources
    for c in columns:
        if c.has_expression:
            # if a column has an expression, it is not a source column
            continue

        ref = c.ref

        if ref.name in sources:
            raise ValueError(f"Duplicate source column name: {ref.name}")
        else:
            col_aliases = {sanitise_column_name(a) for a in c.get_aliases}
            col_aliases.add(sanitise_column_name(ref.name))

            # Check if the alias is already in use
            for a in col_aliases:
                if a in aliases:
                    raise ValueError(
                        f"Duplicate column alias: The sanitized alias `{a}` for column `{ref.name}` already used for `{aliases[a]}`"
                    )
                else:
                    aliases[a] = ref.name

            sources[ref.name] = SourceColDef(
                name=ref.name,
                aliases=col_aliases,
                if_missing=c.if_missing,
                is_meta=c.ref.is_meta,
            )

    return sources


def _reset_registry(nodes: List[ColumnNode], **kwargs):
    return ColumnRegistry([replace(n, **kwargs) for n in nodes])


def _parse_columns_into_list(
    *args: ColumnSchema | Iterable[ColumnSchema] | dict[str, ColumnSchema],
    **kwargs: ColumnSchema,
) -> List[ColumnSchema]:
    cols = []
    # Process positional arguments
    for arg in args:
        if arg is None:
            continue
        if isinstance(arg, dict):
            # Handle dict input like {a: z.col, b: z.col}
            cols.extend(c.with_name(k) for k, c in arg.items())
        elif isinstance(arg, list):
            # Handle list input like [z.col("a"), z.col("b")]
            cols.extend(arg)
        else:
            # Handle single ColumnSchema like z.col("a")
            cols.append(arg)

    # Process keyword arguments
    cols.extend(c.with_name(k) for k, c in kwargs.items())

    return cols



from dataclasses import dataclass, field, KW_ONLY, replace
from typing import TYPE_CHECKING, Literal, Optional

from polars import Expr

from ...types.validation.threshold import Threshold, CheckThreshold, CheckLevel
from .validation_node import get_validation_node

if TYPE_CHECKING:
    from ...ref import ColumnRef
    from ...types import ColumnNode

type ThresholdType = Threshold | dict[str, CheckThreshold] | None


@dataclass(frozen=True)
class RowCheckBase:
    """Configuration for a row check"""

    _: KW_ONLY
    exclude_row: bool = False
    alias: str | None = None
    check_on_cleaned: bool = False
    message: str = field(default="")
    thresholds: ThresholdType = None
    debug: Optional[CheckThreshold] = None
    warning: Optional[CheckThreshold] = None
    error: Optional[CheckThreshold] = None
    reject: Optional[CheckThreshold] = None

    _label: str | None = field(init=False, default=None)

    @classmethod
    def method_id(cls) -> str:
        return "__check_base__"

    def __post_init__(self):
        object.__setattr__(self, "thresholds", self._get_threshold())

    def _get_threshold(self) -> Threshold | None:
        # If we are excluding the row, we want to reject when all rows fail
        default_reject: Literal["all"] | None = (
            "all" if self.exclude_row is True else None
        )

        # self.thresholds takes precedence over self.warning, self.error, self.reject
        if self.thresholds is None:
            if (
                self.debug is not None
                or self.warning is not None
                or self.error is not None
                or self.reject is not None
            ):
                return Threshold(
                    debug=self.debug,
                    warning=self.warning,
                    error=self.error,
                    reject=self.reject or default_reject,
                )

            if self.exclude_row is True:
                return Threshold(warning=True, reject=default_reject)
            else:
                return None

        elif isinstance(self.thresholds, Threshold):
            return Threshold(
                debug=self.thresholds.debug,
                warning=self.thresholds.warning,
                error=self.thresholds.error,
                reject=self.thresholds.reject or default_reject,
            )
        elif isinstance(self.thresholds, dict):
            # invalid_keys = set(self.thresholds.keys()) - {"warning", "error", "reject"}
            # if invalid_keys:
            #     raise ValueError(f"Invalid keys in thresholds: {invalid_keys}")
            return Threshold(
                debug=self.thresholds.get("debug", None),
                warning=self.thresholds.get("warning", None),
                error=self.thresholds.get("error", None),
                reject=self.thresholds.get("reject", default_reject),
            )
        else:
            raise ValueError(f"Invalid type for thresholds: {type(self.thresholds)}")

    def validation_thresholds(
        self,
        debug: Optional[CheckThreshold] = None,
        warning: Optional[CheckThreshold] = None,
        error: Optional[CheckThreshold] = None,
        reject: Optional[CheckThreshold] = None,
        thresholds: Optional[ThresholdType] = None,
        message: Optional[str] = None,
    ):
        """Configure thresholds for this check"""
        if (
            (thresholds is None)
            and (debug is None)
            and (warning is None)
            and (error is None)
            and (reject is None)
        ):
            raise ValueError(
                "At least one of `thresholds`, `debug`, `warning`, `error`, or `reject` must be provided"
            )

        new_thresholds = (
            thresholds
            if thresholds is not None
            else Threshold(debug=debug, warning=warning, error=error, reject=reject)
        )

        return replace(self, thresholds=new_thresholds, message=message or self.message)

    def get_validation_node(self, source_column: "ColumnRef") -> "ColumnNode":
        """Create a validation node for this check"""
        return get_validation_node(
            check=self,
            source_column=source_column,
            thresholds=self._get_threshold(),
            check_on_cleaned=self.check_on_cleaned,
            alias=self.alias,
            label=self._label,
        )

    def _check_expression(
        self, source_column: str, alias: str, level: CheckLevel
    ) -> Expr:
        """Check expression for this check"""
        raise NotImplementedError("Must be implemented in subclass")



from dataclasses import dataclass, field, KW_ONLY
from typing import Any
import polars as pl

from ...ref import ColumnRef
from ...types.data_type import ColumnDataType
from ...types.sensitivity import Sensitivity
from ..._utils.parse_dates import mega_date_handler
from ..._utils.sanitize import (
    SanitiseLevel,
    SanitiseLevelType,
    sanitise_scalar_string,
    full_sanitise_string_col,
    sanitise_string_col,
)

default_data_types = {
    "string": pl.String,
    "number": pl.Int64,
    "boolean": pl.Boolean,
    "date": pl.Date,
    "datetime": pl.Datetime,
    "duration": pl.Duration,
}


def _get_column_names(source: "ColumnRef") -> tuple[str, str]:
    source_column = source.name
    # alias = self.alias if self.alias is not None else source.clean().name
    alias = source.clean().name
    return source_column, alias


@dataclass(frozen=True)
class CleanColumn:
    _: KW_ONLY
    data_type: ColumnDataType = field(default="string")
    col_sensitivity: Sensitivity | None = field(default=None)

    # alias: str | None = field(default=None)

    def clean_ref(self, source: "ColumnRef") -> "ColumnRef":
        return source.clean()

    def clean_expr(self, source: "ColumnRef") -> pl.Expr:
        source_column, alias = _get_column_names(source)

        return pl.col(source_column).cast(pl.String).alias(alias)


@dataclass(frozen=True)
class StringColumn(CleanColumn):
    _: KW_ONLY
    sanitize: SanitiseLevelType = field(default=None)
    sanitise_join_char: str = field(default="_")

    def __post_init__(self):
        if self.sanitize is not None:
            assert self.sanitize in SanitiseLevel, (
                f"sanitize must be one of: {', '.join(f"'{s}'" for s in SanitiseLevel)} or None"
            )

    def _sanitise(self, source_column: str) -> pl.Expr:
        return sanitise_string_col(
            source_column,
            sanitize_level=self.sanitize,
            join_char=self.sanitise_join_char,
        )

    def clean_expr(self, source: "ColumnRef") -> pl.Expr:
        source_column, alias = _get_column_names(source)
        return self._sanitise(source_column).alias(alias)


# @dataclass(frozen=True)
# class SanitisedStringColumn(StringColumn):
#     data_type = "sanitised_string"
#     _: KW_ONLY
#     sanitise_join_char: str = "_"

#     def clean(self) -> pl.Expr:
#         return sanitise_string_col(self._source,  alias=self._alias, join_char=self.sanitise_join_char,)


@dataclass(frozen=True)
class CleanNumberColumn(CleanColumn):
    data_type = "number"
    _: KW_ONLY
    output_format: type[pl.Int64 | pl.Float64] = field(default=pl.Int64)

    def clean_expr(self, source: "ColumnRef") -> pl.Expr:
        source_column, alias = _get_column_names(source)
        return pl.col(source_column).cast(self.output_format, strict=False).alias(alias)


@dataclass(frozen=True)
class CleanBooleanColumn(CleanColumn):
    data_type = "boolean"
    _: KW_ONLY
    true_values: frozenset[Any] = field(default=frozenset({"yes", "y", "true", "1"}))
    false_values: frozenset[Any] = field(default=frozenset({"no", "n", "false", "0"}))

    _bool_map: dict[str, bool] = field(init=False)

    def __post_init__(self):
        if self.true_values is not None:
            assert isinstance(self.true_values, (set, frozenset)), (
                "true_values must be a set or frozenset"
            )
        if self.false_values is not None:
            assert isinstance(self.false_values, (set, frozenset)), (
                "false_values must be a set or frozenset"
            )

        sanitized_true = frozenset(
            sanitise_scalar_string(val) for val in self.true_values
        )
        sanitized_false = frozenset(
            sanitise_scalar_string(val) for val in self.false_values
        )

        if any(value in sanitized_false for value in sanitized_true):
            raise ValueError(
                f"true_values and false_values must be disjoint sets: {sanitized_true} and {sanitized_false}"
            )

        object.__setattr__(
            self,
            "_bool_map",
            {
                **{val: True for val in sanitized_true},
                **{val: False for val in sanitized_false},
            },
        )

    def clean_expr(self, source: "ColumnRef") -> pl.Expr:
        source_column, alias = _get_column_names(source)
        return (
            full_sanitise_string_col(source_column)
            .replace(self._bool_map, default=None, return_dtype=pl.Boolean)
            .alias(alias)
        )


@dataclass(frozen=True)
class CleanDateColumn(CleanColumn):
    data_type = "date"
    _: KW_ONLY
    output_format: pl.Date | pl.Datetime = field(default=pl.Date)

    def __post_init__(self):
        if self.output_format not in [pl.Date, pl.Datetime]:
            raise ValueError("output_format must be either pl.Date or pl.Datetime")

        if self.output_format == pl.Datetime:
            object.__setattr__(self, "data_type", "datetime")

    def clean_expr(self, source: "ColumnRef") -> pl.Expr:
        source_column, alias = _get_column_names(source)
        return mega_date_handler(
            source_column, alias=alias, output_format=self.output_format
        )


@dataclass(frozen=True)
class CleanEnumColumn(StringColumn):
    data_type = "enum"
    _: KW_ONLY
    enum_map: dict[str, Any] = field(default=None)
    sanitize: SanitiseLevelType = field(default="full")
    invalid_value: str | None = "Invalid Value"
    null_value: str | None = "No Data"

    def __post_init__(self):
        assert isinstance(self.enum_map, dict), "enum_map must be a dictionary"

        # if we are sanitizing the column, we need to make sure the enum keys are also sanitized
        object.__setattr__(
            self,
            "enum_map",
            _clean_enum_keys(
                self.enum_map,
                sanitize=self.sanitize,
                sanitise_join_char=self.sanitise_join_char,
            ),
        )

    def clean_expr(self, source: "ColumnRef") -> pl.Expr:
        source_column, alias = _get_column_names(source)
        unique_values = set(self.enum_map.values())
        enum_mapping = self.enum_map

        if self.invalid_value is not None:
            unique_values.add(self.invalid_value)
        if self.null_value is not None:
            unique_values.add(self.null_value)
            enum_mapping = {**enum_mapping, None: self.null_value}


        return (
            self._sanitise(source_column)
            .replace(
                enum_mapping,
                default=self.invalid_value,
                return_dtype=pl.Enum(list(unique_values)),
            )
            .alias(alias)
        )


@dataclass(frozen=True)
class CleanIdColumn(StringColumn):
    data_type = "id"
    _: KW_ONLY
    prefix: ColumnRef | str | None = None
    separator: str = "::"

    def clean_expr(self, source: "ColumnRef") -> pl.Expr:
        source_column, alias = _get_column_names(source)

        if self.prefix is not None:
            prefix_expr = (
                pl.col(self.prefix.name)
                .cast(pl.String)
                .str.strip_chars()
                .str.to_lowercase()
                if isinstance(self.prefix, ColumnRef)
                else pl.lit(str(self.prefix).strip().lower())
            )
            return (prefix_expr + self.separator + self._sanitise(source_column)).alias(
                alias
            )
        else:
            return self._sanitise(source_column).alias(alias)


type CleanStage = (
    CleanColumn
    | StringColumn
    | CleanNumberColumn
    | CleanBooleanColumn
    | CleanDateColumn
    | CleanEnumColumn
    | CleanIdColumn
)


# ---------------------------------------------------------------------------------------------------


def _sanitize_and_validate_keys(
    enum_map: dict[str, Any], sanitize_fn: callable
) -> dict[str, str]:
    sanitized_map = {}
    for k, v in enum_map.items():
        sanitized_key = sanitize_fn(k)
        if sanitized_key in sanitized_map and sanitized_map[sanitized_key] != v:
            raise ValueError(
                f"Duplicate sanitized keys found: '{k}' and '{list(enum_map.keys())[list(sanitized_map.keys()).index(sanitized_key)]}' both sanitize to '{sanitized_key}' but have different values: '{v}' and '{sanitized_map[sanitized_key]}'"
            )
        sanitized_map[sanitized_key] = v
    return sanitized_map


def _clean_enum_keys(
    enum_map: dict[str, Any], *, sanitize: SanitiseLevelType, sanitise_join_char: str
) -> dict[str, str]:
    if sanitize == SanitiseLevel.FULL:
        return _sanitize_and_validate_keys(
            enum_map, lambda k: sanitise_scalar_string(k, join_char=sanitise_join_char)
        )
    elif sanitize == SanitiseLevel.LOWERCASE:
        return _sanitize_and_validate_keys(enum_map, lambda k: k.lower().strip())
    elif sanitize == SanitiseLevel.TRIM:
        return _sanitize_and_validate_keys(enum_map, lambda k: k.strip())
    else:
        return enum_map



from typing import TypedDict, List


class Node(TypedDict):
    id: str
    name: str
    type: str
    schema: str
    stage: str
    validation_level: str | None
    node_category: str
    sensitivity: str


class Edge(TypedDict):
    source: str
    target: str


class Graph(TypedDict):
    nodes: List[Node]
    edges: List[Edge]



from dataclasses import dataclass
from typing import Optional, List, Tuple

COLUMN_NAME_PART_DIVIDER = "__"
COLUMN_NAME_SECTION_DIVIDER = "___"

ID_PART_DIVIDER = "::"
ID_SECTION_DIVIDER = ":::"


@dataclass(frozen=True)
class ColumnIdConfig:
    """Configuration for column IDs"""
    base_name: str
    stage: Optional[str] = None
    schema: Optional[str] = None
    is_clean: bool = False
    is_derived: bool = False
    is_check: bool = False
    is_meta: bool = False
    is_custom_check: bool = False
    check_name: Optional[str] = None

def make_column_id(
    name: str,
    *,
    schema: str,
    is_clean: bool = False,
    is_derived: bool = False,
    is_meta: bool = False,
    is_custom_check: bool = False,
    check_name: Optional[str] = None,
    stage: Optional[str] = None,
    alias_name: Optional[str] = None,
    include_schema_prefix: bool = True,
) -> str:
    base_name = parse_column_name_to_config(name).base_name

    schema_parts = []
    if stage and include_schema_prefix:
        schema_parts.append(stage)
    if schema and include_schema_prefix:
        schema_parts.append(schema)

    # Build prefix
    prefix_parts = []
    if is_meta:
        prefix_parts.append("meta")
    if is_derived:
        prefix_parts.append("derived")
    if is_custom_check:
        prefix_parts.append("custom_check")

    # Build type parts
    type_parts = []
    if is_clean:
        type_parts.append("clean")
    if check_name:
        type_parts.append("check")
        type_parts.append(check_name)

    schema_prefix = [ID_PART_DIVIDER.join(schema_parts)] if len(schema_parts) > 0 else []
    attr_prefix = [ID_PART_DIVIDER.join(prefix_parts)] if len(prefix_parts) > 0 else []

    attr_suffix = [ID_PART_DIVIDER.join(type_parts)] if len(type_parts) > 0 else []

    column_id = ID_SECTION_DIVIDER.join([*schema_prefix, *attr_prefix, base_name, *attr_suffix])

    # Add alias if specified
    if alias_name:
        column_id = f"{column_id}{ID_SECTION_DIVIDER}alias{ID_PART_DIVIDER}{alias_name}"

    return column_id


def parse_column_name_to_config(col_name: str) -> ColumnIdConfig:
    """
    Parse a column name string into a ColumnIdConfig

    Args:
        col_name: The column name string to parse

    Returns:
        ColumnIdConfig
    """
    # If no section divider, treat as simple base name
    if COLUMN_NAME_SECTION_DIVIDER not in col_name:
        return ColumnIdConfig(
            base_name=col_name,
        )

    # First split by section divider to get main parts
    sections = col_name.split(COLUMN_NAME_SECTION_DIVIDER)

    # Initialize variables
    base_name = None
    is_clean = False
    is_derived = False
    is_meta = False
    is_custom_check = False
    check_name = None

    # Valid prefixes and suffixes
    VALID_PREFIXES = {"meta", "derived", "custom_check"}
    VALID_SUFFIXES = {"clean", "check"}

    # Process each section
    for i, section in enumerate(sections):
        if not section:
            continue

        parts = section.split(COLUMN_NAME_PART_DIVIDER)

        # First section: check for valid prefixes
        if i == 0:
            # Check if all parts are valid prefixes
            all_valid_prefixes = all(part in VALID_PREFIXES for part in parts)
            if all_valid_prefixes:
                # Set flags for each valid prefix
                for part in parts:
                    if part == "meta":
                        is_meta = True
                    elif part == "derived":
                        is_derived = True
                    elif part == "custom_check":
                        is_custom_check = True
            else:
                # If any part is not a valid prefix, treat whole section as base name
                base_name = section
            continue

        # Last section: check for valid suffixes
        if i == len(sections) - 1:
            # First check if all parts are valid suffixes or their values
            is_valid_suffix_section = True
            for j, part in enumerate(parts):
                if part not in VALID_SUFFIXES and (j == 0 or parts[j - 1] != "check"):
                    is_valid_suffix_section = False
                    break

            if is_valid_suffix_section:
                # Process valid suffixes
                for j, part in enumerate(parts):
                    if part == "check" and j + 1 < len(parts):
                        check_name = parts[j + 1]
                    elif part == "clean":
                        is_clean = True
            else:
                # If any non-suffix parts found, the whole section is base name
                if base_name is None:
                    base_name = section
                else:
                    base_name = f"{base_name}{COLUMN_NAME_SECTION_DIVIDER}{section}"
            continue

        # Everything else is part of the base name
        if base_name is None:
            base_name = section
        else:
            base_name = f"{base_name}{COLUMN_NAME_SECTION_DIVIDER}{section}"

    return ColumnIdConfig(
        base_name=base_name,
        is_clean=is_clean,
        is_derived=is_derived,
        is_meta=is_meta,
        is_custom_check=is_custom_check,
        check_name=check_name,
    )



from typing import Optional

from ._reference import ColumnRef, parse_column_name

__all__ = [
    "ref",
    "ref_meta",
    "ref_derived",
    "ref_custom_check",
    "ColumnRef",
    "parse_column_name",
]


# -----------------------------------------------------------------------------------------------------------
# Column References
# -----------------------------------------------------------------------------------------------------------


class Ref:
    def __call__(
        self,
        name: str,
        *,
        schema: Optional[str] = None,
        stage: Optional[str] = None,
        is_clean: bool = False,
        is_derived: bool = False,
        is_check: bool = False,
        is_meta: bool = False,
        is_custom_check: bool = False,
        check_name: str | None = None,
    ) -> ColumnRef:
        """
        Create a column reference for use in schema definitions.

        Parameters:
            name: Name of the column.

            is_clean: Whether the column is cleaned.
            is_derived: Whether the column is derived (calculated from other columns).
            is_check: Whether the column is a check/validation.
            is_meta: Whether the column is a meta column.
            is_custom_check: Whether the column is a custom check derived from an expression.
            check_name: Name of the check/validation.
            schema: Table schema name.
            stage: Pipeline stage name.

        Returns:
            ColumnRef: The column reference.
        """
        parsed = parse_column_name(name)

        return ColumnRef(
            base_name=parsed.base_name,
            schema=schema,
            stage=stage,
            is_clean=is_clean or parsed.is_clean,
            is_derived=is_derived or parsed.is_derived,
            is_check=is_check or parsed.is_check,
            is_meta=is_meta or parsed.is_meta,
            is_custom_check=is_custom_check or parsed.is_custom_check,
            check_name=check_name or parsed.check_name,
        )

    def __getattr__(self, name: str) -> ColumnRef:
        # For autocomplete to work with IPython
        if name.startswith("__wrapped__"):
            return getattr(type(self), name)

        return parse_column_name(name)


ref: Ref = Ref()
"""
   Create a column reference for use in schema definitions.
   
   Example:
    zeolite.ref("other_column")

   Parameters:
       name: Name of the column.

       is_clean: Whether the column is cleaned.
       is_derived: Whether the column is derived (calculated from other columns).
       is_check: Whether the column is a check/validation.
       is_meta: Whether the column is a meta column.
       is_custom_check: Whether the column is a custom check derived from an expression.
       check_name: Name of the check/validation.
       schema: Table schema name.
       stage: Pipeline stage name.

   Returns:
       ColumnRef: The column reference.
   """


# -----------------------------------------------------------------------------------------------------------
# Column Reference Shortcuts
# -----------------------------------------------------------------------------------------------------------
def ref_meta(name: str) -> ColumnRef:
    """
    Create a column reference for a meta column.

    Example:
    zeolite.ref_meta("other_column")

    Parameters:
        name: Name of the column.

    Returns:
        ColumnRef: The column reference.
    """
    return ref(name=name, is_meta=True)


def ref_derived(name: str) -> ColumnRef:
    """
    Create a column reference to a derived column whose
    value is computed from an expression.

    Example:
    zeolite.ref_derived("other_column")

    Parameters:
        name: Name of the column.

    Returns:
        ColumnRef: The column reference.
    """
    return ref(name=name, is_derived=True)


def ref_custom_check(name: str) -> ColumnRef:
    """
    Create a column reference to derived custom check/validation
    (that is computed from an expression).

    Example:
    zeolite.ref_custom_check("other_column")

    Parameters:
        name: Name of the column.

    Returns:
        ColumnRef: The column reference.
    """
    return ref(name=name, is_custom_check=True, is_derived=True)



from dataclasses import dataclass, field, KW_ONLY
from enum import Enum

import polars as pl

from .check_base import RowCheckBase
from .data_checks import (
    ROW_VALIDATION_SUCCESS_VALUE,
    check_col_row_is_unique,
    check_col_row_not_empty,
)
from ...types.validation.threshold import CheckLevel


class RowCheckType(Enum):
    IS_EMPTY = "is_empty"
    IS_DUPLICATED = "is_duplicated"
    IS_INVALID_DATE = "is_invalid_date"
    IS_EQUAL_TO = "is_equal_to"


@dataclass(frozen=True)
class IsValueEmpty(RowCheckBase):
    """
    Validation check: Ensures that column values are not empty or null.

    Parameters:
        exclude_row (bool): Whether to exclude rows with empty values.
        check_on_cleaned (bool): Whether to check on cleaned column or the original.

        thresholds (Threshold): Thresholds for error capturing when the table is processed.
        warning (CheckThreshold): Warning threshold (used when no thresholds are provided).
        error (CheckThreshold): Error threshold (used when no thresholds are provided).
        reject (CheckThreshold): Reject threshold (used when no thresholds are provided).

        treat_empty_strings_as_null (bool): Whether to treat empty strings as null.
        message (str): Error message template.
    """

    @classmethod
    def method_id(cls) -> str:
        return RowCheckType.IS_EMPTY.value

    treat_empty_strings_as_null: bool = False
    message: str = field(
        default="{{column}} has {{count}} null (empty) value(s) ({{fraction}})"
    )

    def _check_expression(
        self, source_column: str, alias: str, level: CheckLevel
    ) -> pl.Expr:
        return check_col_row_not_empty(
            source_column,
            alias=alias,
            str_check=self.treat_empty_strings_as_null,
            error_as_value=level,
        )


@dataclass(frozen=True)
class IsValueDuplicated(RowCheckBase):
    """
    Validation check: Ensures that column values are unique.

    Parameters:
    exclude_row (bool): Whether to exclude rows with empty values.
        check_on_cleaned (bool): Whether to check on cleaned column or the original.
        thresholds (Threshold): Thresholds for error capturing when the table is processed.
        warning (CheckThreshold): Warning threshold (used when no thresholds are provided).
        error (CheckThreshold): Error threshold (used when no thresholds are provided).
        reject (CheckThreshold): Reject threshold (used when no thresholds are provided).
        message (str): Error message template.
    """

    @classmethod
    def method_id(cls) -> str:
        return RowCheckType.IS_DUPLICATED.value

    message: str = field(
        default="{{column}} has {{count}} duplicate value(s) ({{fraction}})"
    )

    def _check_expression(
        self, source_column: str, alias: str, level: CheckLevel
    ) -> pl.Expr:
        return check_col_row_is_unique(source_column, alias=alias, error_as_value=level)


@dataclass(frozen=True)
class IsValueInvalidDate(RowCheckBase):
    """
    Validation check: Ensures that column values are valid dates - essentially checks
    for nulls in a date data-type column.

    Parameters:
        exclude_row (bool): Whether to exclude rows with empty values.
        check_on_cleaned (bool): Whether to check on cleaned column or the original.
        thresholds (Threshold): Thresholds for error capturing when the table is processed.
        warning (CheckThreshold): Warning threshold (used when no thresholds are provided).
        error (CheckThreshold): Error threshold (used when no thresholds are provided).
        reject (CheckThreshold): Reject threshold (used when no thresholds are provided).
        message (str): Error message template.
    """

    @classmethod
    def method_id(cls) -> str:
        return RowCheckType.IS_INVALID_DATE.value

    check_on_cleaned: bool = True
    message: str = field(
        default="{{column}} has {{count}} invalid date(s) ({{fraction}})"
    )

    def _check_expression(
        self, source_column: str, alias: str, level: CheckLevel
    ) -> pl.Expr:
        return check_col_row_not_empty(
            source_column, alias=alias, str_check=False, error_as_value=level
        )


@dataclass(frozen=True)
class IsValueEqualTo(RowCheckBase):
    """
    Validation check: Ensures that column values are (not) equal to a specified value. This will
    throw an error if the value IS FOUND in the column.

    Parameters:
        value (str | int | float | bool): The value to check against.
        exclude_row (bool): Whether to exclude rows with empty values.
        check_on_cleaned (bool): Whether to check on cleaned column or the original.
        thresholds (Threshold): Thresholds for error capturing when the table is processed.
        warning (CheckThreshold): Warning threshold (used when no thresholds are provided).
        error (CheckThreshold): Error threshold (used when no thresholds are provided).
        reject (CheckThreshold): Reject threshold (used when no thresholds are provided).
        message (str): Error message template.
    """

    @classmethod
    def method_id(cls) -> str:
        return RowCheckType.IS_EQUAL_TO.value

    value: str | int | float | bool = None
    _: KW_ONLY
    check_on_cleaned: bool = False
    message: str = field(
        default="{{column}} has {{count}} value(s) equal to '{{match_value}}' ({{fraction}})"
    )

    def __post_init__(self):
        assert self.value is not None, "Must provide a value to check against"
        object.__setattr__(
            self, "_label", f"{self.method_id()}_{self.value}".lower().replace(" ", "_")
        )
        object.__setattr__(
            self, "message", self.message.replace("{{match_value}}", str(self.value))
        )
        super().__post_init__()

    def _check_expression(
        self, source_column: str, alias: str, level: CheckLevel
    ) -> pl.Expr:
        return (
            pl.when(pl.col(source_column) == self.value)
            .then(pl.lit(level))
            .otherwise(pl.lit(ROW_VALIDATION_SUCCESS_VALUE))
            .alias(alias)
        )


type ColumnCheckType = (
    IsValueEmpty | IsValueDuplicated | IsValueInvalidDate | IsValueEqualTo
)



from dataclasses import field
from typing import List, Any

from polars import Expr, lit

from ..ref import ColumnRef
from ..types import ThresholdLevel
from ..types.data_type import ColumnDataType
from ..types.sensitivity import Sensitivity
from ._base import ColumnSchema
from ._clean import CleanStage
from .validation import ColumnCheckType
from .validation.check_base import ThresholdType

__all__ = [
    "col",
    "column",
    "ColumnSchema",
    "str_col",
    "bool_col",
    "date_col",
    "int_col",
    "float_col",
    "derived_col",
    "derived_custom_check",
    "meta_col",
]


# -----------------------------------------------------------------------------------------------------------
# Column Definitions
# -----------------------------------------------------------------------------------------------------------
def _create_col(
    ref: ColumnRef,
    data_type: ColumnDataType = "unknown",
    sensitivity: Sensitivity = None,
    aliases: set[str] = None,
    validations: List[ColumnCheckType] = None,
    clean: CleanStage | ColumnDataType | dict[str, Any] | None = None,
) -> ColumnSchema:
    return ColumnSchema(
        col_ref=ref,
        data_type=data_type,
        sensitivity=sensitivity,
        aliases=aliases,
        validations=validations,
        clean=clean,
    )


class Col:
    data_type: ColumnDataType

    def __init__(self, data_type: ColumnDataType = "unknown"):
        self.data_type = data_type

    def __call__(
        self,
        name: str | None = None,
        *,
        data_type: ColumnDataType | None = None,
        sensitivity: Sensitivity = None,
        aliases: set[str] = None,
        validations: List[ColumnCheckType] = None,
        clean: CleanStage | ColumnDataType | dict[str, Any] | None = None,
    ) -> ColumnSchema:
        """
        Define a new column.

        Parameters:
            name: Name of the column.
            data_type: Data type of the column.
            sensitivity: Sensitivity of the column.
            aliases: Aliases for the column.
            validations: List of validation checks.
            clean: Clean stage for the column.

        Returns:
            ColumnSchema: The column schema.
        """
        return _create_col(
            ref=ColumnRef(name),
            data_type=data_type or self.data_type,
            sensitivity=sensitivity,
            aliases=aliases,
            validations=validations,
            clean=clean,
        )

    def __getattr__(self, name: str) -> ColumnSchema:
        # For autocomplete to work with IPython
        if name.startswith("__wrapped__"):
            return getattr(type(self), name)

        return _create_col(ColumnRef(name))


column: Col = Col()
col: Col = Col()
"""
Define a new column.
Example:
    zeolite.col("id").validations(zeolite.check_is_value_empty(), ...)

Parameters:
    name: Name of the column.
    data_type: Data type of the column.
    sensitivity: Sensitivity of the column.
    aliases: Aliases for the column.
    validations: List of validation checks.
    clean: Clean stage for the column.

Returns:
    ColumnSchema: The column schema.
"""


# -----------------------------------------------------------------------------------------------------------
# Data-Type Specific Column Definitions
# -----------------------------------------------------------------------------------------------------------

str_col = Col("string")
"""
Helper to define a new string column.
"""

bool_col = Col("boolean")
"""
Helper to define a new boolean column.
"""

date_col = Col("date")
"""
Helper to define a new date column.
"""

int_col = Col("integer")
"""
Helper to define a new integer column.
"""


float_col = Col("float")
"""
Helper to define a new float column.
"""


# -----------------------------------------------------------------------------------------------------------
# Derived Column Definitions
# -----------------------------------------------------------------------------------------------------------
class DerivedCol:
    def __call__(
        self,
        name: str | None = None,
        *,
        function: Expr,
        data_type: ColumnDataType = "unknown",
        sensitivity: Sensitivity = None,
        validations: List[ColumnCheckType] = None,
    ) -> ColumnSchema:
        """
        Define a derived column whose value is computed from an expression.

        Parameters:
            name: Name of the derived column.
            function: Polars expression to compute the column.
            data_type: (Optional) Data type of the column.
            sensitivity: (Optional) Sensitivity of the column.
            validations: (Optional) List of validation checks.

        Returns:
            ColumnSchema: The derived column schema.
        """
        return _create_col(
            ref=ColumnRef(name).derived(),
            data_type=data_type,
            sensitivity=sensitivity,
            validations=validations,
        ).derived(function)

    def __getattr__(self, name: str) -> ColumnSchema:
        # For autocomplete to work with IPython
        if name.startswith("__wrapped__"):
            return getattr(type(self), name)

        # We add a literal to the expression to ensure that the derived column has an expression
        return _create_col(ColumnRef(name).derived()).derived(lit(name))


derived_col: DerivedCol = DerivedCol()
"""
Define a derived column whose value is computed from an expression.

Parameters:
    name: Name of the derived column.
    function: Polars expression to compute the column.
    data_type: (Optional) Data type of the column.
    sensitivity: (Optional) Sensitivity of the column.
    validations: (Optional) List of validation checks.

Returns:
    ColumnSchema: The derived column schema.
"""

class DerivedCustomCheckCol:
    def __call__(
        self,
        name: str | None = None,
        *,
        function: Expr,
        sensitivity: Sensitivity = Sensitivity.NON_SENSITIVE,
        thresholds: ThresholdType = None,
        message: str = field(default=""),
    ) -> ColumnSchema:
        """
        Define a derived custom check/validation that is computed from an expression.

        Parameters:
            name: Name of the derived validation.
            function: Polars expression to compute the validation.
            sensitivity: (Optional) Sensitivity of the validation.
            thresholds: (Optional) Thresholds for the validation.
            message: (Optional) Message for the validation.

        Returns:
            ColumnSchema: The derived validation schema.
        """
        return _create_col(
            ref=ColumnRef(name).custom_check(),
            data_type="boolean",
            sensitivity=sensitivity,
        ).custom_check(function, thresholds, message)

    def __getattr__(self, name: str) -> ColumnSchema:
        # For autocomplete to work with IPython
        if name.startswith("__wrapped__"):
            return getattr(type(self), name)

        # We add a pass default to ensure that the custom_check has an expression
        return _create_col(ColumnRef(name).custom_check()).custom_check(
            lit(ThresholdLevel.PASS.level), None, "Not implemented"
        )


derived_custom_check: DerivedCustomCheckCol = DerivedCustomCheckCol()
"""
Define a derived custom check/validation that is computed from an expression.

Parameters:
   name: Name of the derived validation.
   function: Polars expression to compute the validation.
   sensitivity: (Optional) Sensitivity of the validation.
   thresholds: (Optional) Thresholds for the validation.
   message: (Optional) Message for the validation.

Returns:
   ColumnSchema: The derived validation schema.
"""

class MetaCol:
    def __call__(
        self,
        name: str | None = None,
        *,
        function: Expr = None,
        data_type: ColumnDataType = "unknown",
        sensitivity: Sensitivity = None,
    ) -> ColumnSchema:
        """
        Define a meta column - this is a special column that is usually added
        to the data e.g. during initial ingestion, and should be identified as
        separate from the source data.

        Parameters:
            name: Name of the meta column.
            function: (Optional) Polars expression to compute the column.
            data_type: (Optional) Data type of the column.
            sensitivity: (Optional) Sensitivity of the column.

        Returns:
            ColumnSchema: The meta column schema.
        """
        m_col = _create_col(
            ref=ColumnRef(name, is_meta=True),
            data_type=data_type,
            sensitivity=sensitivity,
        )
        return m_col.derived(function) if function is not None else m_col

    def __getattr__(self, name: str) -> ColumnSchema:
        # For autocomplete to work with IPython
        if name.startswith("__wrapped__"):
            return getattr(type(self), name)

        return _create_col(ColumnRef(name, is_meta=True))

meta_col: MetaCol = MetaCol()
"""
Define a meta column - this is a special column that is usually added
to the data e.g. during initial ingestion, and should be identified as
separate from the source data.

Parameters:
    name: Name of the meta column.
    function: (Optional) Polars expression to compute the column.
    data_type: (Optional) Data type of the column.
    sensitivity: (Optional) Sensitivity of the column.

Returns:
    ColumnSchema: The meta column schema.
"""


from typing import Union, List, Set, KeysView, Literal

import polars as pl

from ...types import ThresholdLevel
from ..._utils.sanitize import full_sanitise_string_col

ROW_VALIDATION_SUCCESS_VALUE = ThresholdLevel.PASS.level

"""
---------------------------------------------------------------------------------------------
    Validator functions
---------------------------------------------------------------------------------------------
These functions check the column to check it is valid
"""


def _error_as_value(check: pl.Expr, error: str) -> pl.Expr:
    enum_def = pl.Enum([ROW_VALIDATION_SUCCESS_VALUE, error])

    return (
        pl.when(check)
        .then(pl.lit(ROW_VALIDATION_SUCCESS_VALUE))
        .otherwise(pl.lit(error))
        .cast(enum_def, strict=False)
    )


def check_col_row_not_empty(
    col: str,
    *,
    prefix: str = "",
    alias: str = None,
    str_check=False,
    error_as_value: None | str = None,
) -> pl.Expr:
    check = (
        (
            pl.col(col).is_not_null()
            & (full_sanitise_string_col(col).str.len_chars() > 0)
        )
        if str_check
        else pl.col(col).is_not_null()
    )
    a = alias if alias else f"{prefix}is_{col}_not_empty"
    if error_as_value is not None:
        return _error_as_value(check, error_as_value).alias(a)
    else:
        return check.alias(a)


def check_col_row_is_unique(
    col: str, *, prefix: str = "", alias: str = None, error_as_value: None | str = None
) -> pl.Expr:
    a = alias if alias else f"{prefix}is_{col}_unique"
    if error_as_value is not None:
        return _error_as_value(pl.col(col).is_unique(), error_as_value).alias(a)
    else:
        return pl.col(col).is_unique().alias(a)


_EnumMatchesType = Union[List[str], Set[str], KeysView[str]]
_MatchReturnType = Literal["match", "boolean", "error"]


def check_enum_match(
    col: str,
    *,
    prefix: str = "",
    alias: str = None,
    matches: _EnumMatchesType,
    sanitize: bool = True,
    return_type: _MatchReturnType = "match",
    error_value: str = "__NO_MATCH__",
) -> pl.Expr:
    if sanitize:
        return (
            pl.when(pl.col(col).is_null())
            .then(pl.lit(None))
            .when(full_sanitise_string_col(col).is_in(matches))
            .then(
                pl.lit(True)
                if return_type == "boolean"
                else full_sanitise_string_col(col)
            )
            .otherwise(pl.lit(False if return_type == "boolean" else error_value))
            .alias(alias if alias else f"{prefix}{col}_enum_match")
        )

    else:
        return (
            pl.when(pl.col(col).is_null())
            .then(pl.lit(None))
            .when(pl.col(col).is_in(matches))
            .then(pl.lit(True) if return_type == "boolean" else pl.col(col))
            .otherwise(pl.lit(False if return_type == "boolean" else "__NO_MATCH__"))
            .alias(alias if alias else f"{prefix}{col}_enum_match")
        )



from typing import Any, List, Tuple
from collections.abc import Iterable


def flatten_args(
    args: Iterable[Any | Iterable[Any]] | Tuple[Any | List[Any], ...],
) -> List[Any]:
    return [
        item
        for sublist in args
        for item in (sublist if isinstance(sublist, list) else [sublist])
    ]



from collections import defaultdict

from ..types import ColumnNode


# %%
def verify_column_registry_integrity(columns: list[ColumnNode]) -> bool:
    """
    Verify column definitions by checking:
    1. All parent_columns reference existing column names
    2. All parent_ids match existing column IDs
    3. parent_ids are properly mapped from parent_columns

    Args:
        columns: List of column definitions

    Raises:
        ValueError: If validation fails, with detailed error message
    """
    # Create lookup maps
    column_by_id = {col.id: col for col in columns}
    column_by_name = {col.name: col for col in columns}

    # Check source references and mappings
    unmapped_sources = defaultdict(set)
    invalid_parent_ids = defaultdict(set)
    incorrect_mappings = defaultdict(set)

    for col in columns:
        # Check parent columns that don't map to any known column
        unmapped = {src for src in col.parent_columns if src not in column_by_name}
        if unmapped:
            unmapped_sources[col.id].update(unmapped)

        # Check parent_ids that don't exist
        invalid_ids = {
            src_id for src_id in col.parent_ids if src_id not in column_by_id
        }
        if invalid_ids:
            invalid_parent_ids[col.id].update(invalid_ids)

        # Check that all valid parent_columns are properly mapped to parent_ids
        for src in col.parent_columns:
            if src in column_by_name:
                expected_id = column_by_name[src].id
                if expected_id not in col.parent_ids:
                    incorrect_mappings[col.id].add(src)

    # Build error message if any issues found
    errors = []

    if unmapped_sources:
        error_details = "\n".join(
            f"Column '{col_id}' references non-existent columns: {sorted(refs)}"
            for col_id, refs in sorted(unmapped_sources.items())
        )
        errors.append(f"Unmapped source columns found:\n{error_details}")

    if invalid_parent_ids:
        error_details = "\n".join(
            f"Column '{col_id}' contains invalid source IDs: {sorted(refs)}"
            for col_id, refs in sorted(invalid_parent_ids.items())
        )
        errors.append(f"Invalid source IDs found:\n{error_details}")

    if incorrect_mappings:
        error_details = "\n".join(
            f"Column '{col_id}' has unmapped valid sources: {sorted(refs)}"
            for col_id, refs in sorted(incorrect_mappings.items())
        )
        errors.append(f"Incorrect source mappings found:\n{error_details}")

    if errors:
        raise ValueError("\n\n".join(errors))

    return True



from collections import defaultdict
from typing import List

from ..types import ColumnNode


# %%
def _extract_calculated_columns(columns: List[ColumnNode]) -> List[ColumnNode]:
    """
    Filter out columns with None expressions and clean up removed columns from dependencies
    """
    # Identify calculated columns
    columns_with_expr = {col.id for col in columns if col.expression is not None}

    # Remove non-calc columns and clean up calculated dependencies
    filtered_columns = []
    for col in columns:
        if col.expression is not None:
            # Clean up parent_ids to only reference calculated columns
            cleaned_col = col.with_parent_ids(
                src for src in col.parent_ids if src in columns_with_expr
            )
            filtered_columns.append(cleaned_col)

    return filtered_columns


# %%
def generate_optimized_stages(all_columns: List[ColumnNode]) -> List[List[ColumnNode]]:
    """
    Split columns into sequential stages based on their dependencies (using topological search!).
    Each stage contains complete column definitions that can be calculated once all columns
    in previous stages are available.
    """
    columns = _extract_calculated_columns(all_columns)

    # Create a map of column id to its definition for easy lookup
    column_map = {col.id: col for col in columns}

    # Create dependency graph and track in-degrees
    dependencies = defaultdict(set)
    in_degrees = defaultdict(int)

    # Build the dependency graph
    for col in columns:
        col_id = col.id
        for source in col.parent_ids:
            if (
                source in column_map
            ):  # Only count dependencies that exist in our column list
                dependencies[source].add(col_id)
                in_degrees[col_id] += 1

    # Initialize stages
    stages = []

    # Find all columns with no dependencies (in_degree = 0)
    zero_degree = [col.id for col in columns if in_degrees[col.id] == 0]

    while zero_degree:
        # Process current level
        current_stage = []
        next_zero_degree = []

        # Add all current zero-degree nodes to current stage
        for col_id in zero_degree:
            current_stage.append(column_map[col_id])

            # Reduce in_degree for all dependent columns
            for dependent in dependencies[col_id]:
                in_degrees[dependent] -= 1
                if in_degrees[dependent] == 0:
                    next_zero_degree.append(dependent)

        stages.append(current_stage)
        zero_degree = next_zero_degree

    # Check for circular dependencies
    remaining_columns = sum(1 for col in columns if in_degrees[col.id] > 0)
    if remaining_columns > 0:
        raise ValueError(
            f"Circular dependency detected! {remaining_columns} columns could not be processed."
        )

    return stages



from dataclasses import dataclass, field, replace
from typing import Optional, Literal
from polars import Expr, LazyFrame

from .data_type import ColumnDataType
from .sensitivity import Sensitivity
from .validation.threshold import ThresholdLevel
from .validation.validation_rule import ColumnValidationRule
from .error import (
    SchemaValidationError,
    FileValidationError,
    StructureValidationError,
    DataValidationError,
    UnknownValidationError,
)

__all__ = [
    "ColumnNode",
    "SourceColDef",
    "ValidationResult",
    "ProcessingResult",
    "ProcessingFailure",
    "ProcessingSuccess",
    "SchemaValidationError",
    "FileValidationError",
    "StructureValidationError",
    "DataValidationError",
    "UnknownValidationError",
    "ThresholdLevel",
]


@dataclass(frozen=True)
class ColumnNode:
    id: str
    name: str
    data_type: ColumnDataType
    column_type: Literal[
        "source", "cleaned", "validation", "meta", "derived", "custom_validation"
    ]
    schema: str
    stage: str
    sensitivity: Sensitivity
    expression: Optional[Expr] = None
    validation_rule: Optional[ColumnValidationRule] = None
    parent_columns: frozenset[str] = field(default_factory=frozenset)
    parent_ids: frozenset[str] = field(default_factory=frozenset)

    def __post_init__(self):
        if self.expression is not None:
            assert self.name == self.expression.meta.output_name(), (
                f"Column name {self.name} does not match expression output name {self.expression.meta.output_name()}"
            )
            if len(self.parent_columns) == 0:
                object.__setattr__(
                    self,
                    "parent_columns",
                    frozenset({c for c in self.expression.meta.root_names()}),
                )

    def with_parent_ids(self, parent_ids: set[str]) -> "ColumnNode":
        return replace(self, parent_ids=frozenset(parent_ids))


@dataclass(frozen=True)
class SourceColDef:
    name: str
    aliases: set[str]
    if_missing: ThresholdLevel
    is_meta: bool


# Validation Results
@dataclass(frozen=True)
class ValidationResult:
    data: LazyFrame
    errors: list[SchemaValidationError]
    reject: bool = False


@dataclass(frozen=True)
class ProcessingResult:
    normalised: LazyFrame | None
    prepared: LazyFrame | None
    validated: LazyFrame | None
    errors: list[SchemaValidationError]
    success: bool = False
    failed_stage: Literal["normalise", "prepare", "validate"] | None = None


@dataclass(frozen=True)
class ProcessingFailure(ProcessingResult):
    success: False = False
    failed_stage: Literal["normalise", "prepare", "validate"]


@dataclass(frozen=True)
class ProcessingSuccess(ProcessingResult):
    normalised: LazyFrame
    prepared: LazyFrame
    validated: LazyFrame
    success: True = True
    failed_stage: None = None



from typing import List

import polars as pl

from ...types import SourceColDef, ThresholdLevel, ValidationResult
from ...types.error import StructureValidationError
from ..._utils.sanitize import sanitise_column_name


# _INTERNAL_COL_PATTERN = r'meta__.*'
# %%


def _rename_columns_from_aliases(
    lf: pl.LazyFrame,
    *,
    col_defs: List[SourceColDef],
    schema_name: str,
    source_name: str | None,
) -> ValidationResult:
    """
     Rename columns in a LazyFrame based on a dictionary of aliases.

    :param lf: Polars LazyFrame
    :param col_defs: Column definitions (from a TableSchema)
    :param schema_name: Name of the schema
    :param source_name: (Optional) Name of the source file (used for error messages)
    :return:
    """
    errors = []

    schema_aliases = {a: c.name for c in col_defs for a in c.aliases}

    original_columns = lf.collect_schema().names()
    original_columns_map = {(c, sanitise_column_name(c)) for c in original_columns}

    assigned_targets = set()
    rename_cols = {}

    for original_col, sanitized_original in original_columns_map:
        target = schema_aliases.get(sanitized_original, None)

        # print(original_col, " -> ", sanitized_original, " -> ", target, " -> ", target not in assigned_targets)
        if target is None:
            continue

        # Only proceed if source & target aren't the same
        if target != original_col:
            # If the target is already assigned, or present in the original columns, then this is a duplicate
            if target in assigned_targets or (
                target in original_columns and target != original_col
            ):
                errors.append(
                    StructureValidationError(
                        schema=schema_name,
                        source=source_name,
                        # column=original_columns_map[sanitized_original],
                        column=original_col,
                        error="duplicate_column",
                        level=ThresholdLevel.REJECT.level,
                        message=f"Column `{original_col}` cannot be assigned, as another column already maps to `{target}`",
                    )
                )
            else:
                rename_cols[original_col] = target

        assigned_targets.add(target)

    for old_col, new_col in rename_cols.items():
        if old_col != new_col:
            errors.append(
                StructureValidationError(
                    schema=schema_name,
                    source=source_name,
                    column=old_col,
                    error="renamed_column",
                    level=ThresholdLevel.DEBUG.level,
                    message=f"Renamed `{old_col}` to `{new_col}`",
                )
            )

    return ValidationResult(data=lf.rename(rename_cols), errors=errors)


# %%


def _remove_null_rows(lf: pl.LazyFrame) -> pl.LazyFrame:
    return lf.filter(~pl.all_horizontal(pl.all().is_null()))


# %%


def _check_critical_columns(
    data_cols: list[str],
    *,
    col_defs: List[SourceColDef],
    schema_name: str,
    source_name: str | None,
) -> list[StructureValidationError]:
    critical_cols = {
        c.name: c.if_missing for c in col_defs if c.if_missing == ThresholdLevel.REJECT
    }
    return [
        StructureValidationError(
            schema=schema_name,
            source=source_name,
            column=col_key,
            error="required_column",
            level=lvl.level,
            message=f"Column '{col_key}' is required but missing from  from the data",
        )
        for col_key, lvl in critical_cols.items()
        if col_key not in data_cols
    ]


def _check_non_critical_columns(
    data_cols: list[str],
    *,
    col_defs: List[SourceColDef],
    schema_name: str,
    source_name: str | None,
) -> list[StructureValidationError]:
    non_critical_cols = {
        c.name: c.if_missing
        for c in col_defs
        if c.if_missing != ThresholdLevel.REJECT and c.if_missing is not None
    }
    return [
        StructureValidationError(
            schema=schema_name,
            source=source_name,
            column=col_key,
            error="missing_column",
            level=lvl.level,
            message=f"Column '{col_key}' is missing from the data",
        )
        for col_key, lvl in non_critical_cols.items()
        if col_key not in data_cols
    ]


def _check_extra_columns(
    data_cols: list[str],
    *,
    col_defs: List[SourceColDef],
    schema_name: str,
    source_name: str | None,
) -> list[StructureValidationError]:
    all_cols = {c.name for c in col_defs}
    return [
        StructureValidationError(
            schema=schema_name,
            source=source_name,
            column=c,
            error="extra_column",
            level=ThresholdLevel.DEBUG.level,
            message=f"Column '{c}' is additional and not required",
        )
        for c in data_cols
        if c not in all_cols  # and not re.match(_INTERNAL_COL_PATTERN, c)
    ]


# %%


def _normalise_table_structure(
    lf: pl.LazyFrame,
    col_defs: List[SourceColDef],
    schema_name: str,
    source_name: str | None,
) -> ValidationResult:
    """
    Normalise the structure of a table based on the column definitions.

    :param lf: Polars LazyFrame
    :param schema_name: The name of the schema
    :param col_defs: List of Column Schema definitions
    :param source_name: (Optional) Name of the source file (used for error messages)
    :return:
    """
    errors: list[StructureValidationError] = []
    missing_cols: list[StructureValidationError] = []
    data_cols = lf.collect_schema().names()

    critical_errors = _check_critical_columns(
        data_cols, col_defs=col_defs, schema_name=schema_name, source_name=source_name
    )
    missing_cols.extend(critical_errors)

    non_critical = _check_non_critical_columns(
        data_cols, col_defs=col_defs, schema_name=schema_name, source_name=source_name
    )
    missing_cols.extend(non_critical)

    if len(missing_cols) > 0:
        lf = lf.with_columns(
            [pl.lit(None).cast(pl.String).alias(mc.column) for mc in missing_cols]
        )
        errors.extend(missing_cols)

    extra_cols = _check_extra_columns(
        data_cols, col_defs=col_defs, schema_name=schema_name, source_name=source_name
    )

    # if lf has extra columns, drop them
    if len(extra_cols) > 0:
        lf = lf.drop([c.column for c in extra_cols])
        errors.extend(extra_cols)

    is_empty = (
        lf.select([c.name for c in col_defs if not c.is_meta])
        .pipe(_remove_null_rows)
        .limit(1)
        .collect()
        .is_empty()
    )
    if is_empty:
        errors.append(
            StructureValidationError(
                schema=schema_name,
                source=source_name,
                error="empty_data",
                level=ThresholdLevel.REJECT.level,
                message=f"`{schema_name}` has no data after additional/unused columns have been removed",
            )
        )

    return ValidationResult(data=lf, errors=errors)


# %%
def normalise_column_headers(
    lf: pl.LazyFrame,
    col_defs: List[SourceColDef],
    schema_name: str,
    source_name: str | None,
) -> ValidationResult:
    """
    Normalise the column headers of a table to a common structure based on the column definitions.
    :param lf: Polars LazyFrame
    :param col_defs: Column definitions (from a TableSchema)
    :param schema_name: Name of the schema
    :param source_name: (Optional) Name of the source file (used for error messages)
    :return:
    """
    errors = []
    reject = False

    sanitised = _rename_columns_from_aliases(
        lf, col_defs=col_defs, schema_name=schema_name, source_name=source_name
    )
    errors.extend(sanitised.errors)

    normalised = _normalise_table_structure(
        sanitised.data,
        col_defs=col_defs,
        schema_name=schema_name,
        source_name=source_name,
    )
    errors.extend(normalised.errors)

    for e in errors:
        if e.level == ThresholdLevel.REJECT.level:
            reject = True
            break

    return ValidationResult(data=normalised.data, errors=errors, reject=reject)



from enum import StrEnum


class Sensitivity(StrEnum):
    """
    Represents a classification system for sensitivity levels of data.

    Attributes:
        HIGHLY_SENSITIVE: Represents highly sensitive data, such as private or
            critical information.
        CONFIDENTIAL: Represents confidential data that requires restricted
            access.
        IDENTIFIABLE: Represents data that can directly identify individuals.
        POTENTIALLY_IDENTIFIABLE: Represents data that may not directly identify
            an individual but could do so when combined with other information.
        LIMITED: Operational or administrative data with minimal personal
            information but still requires protection.
        NON_SENSITIVE: Non-identifiable metadata or system information that
            poses minimal privacy risk
        UNKNOWN: Represents data with an unknown or undefined level of
            sensitivity.
    """

    HIGHLY_SENSITIVE = "1_highly_sensitive"
    CONFIDENTIAL = "2_confidential"
    IDENTIFIABLE = "3_identifiable"
    POTENTIALLY_IDENTIFIABLE = "4_potentially_identifiable"
    LIMITED = "5_limited"
    NON_SENSITIVE = "6_non_sensitive"
    UNKNOWN = "99_unknown"



from ._table import TableSchema


schema = TableSchema
"""
Defines a table schema for data validation and processing.

Parameters:
    name (str): Name of the schema.
    columns (List[ColumnSchema] | dict[str, ColumnSchema], optional): List of column schemas.
    is_required (bool): Whether the schema is required.
    stage (str, optional): Processing stage.

Example:
    demo_schema = z.schema("demo", columns=[...])
"""


__all__ = ["TableSchema", "schema"]



import pytest
import polars as pl
from zeolite.registry import ColumnRegistry
from zeolite.types import ColumnNode, Sensitivity

common_params = {
    "sensitivity": Sensitivity.NON_SENSITIVE,
    "schema": "test",
    "stage": "test_stage",
}

base_column_params = {
    "data_type": "string",
    "column_type": "source",
    **common_params,
}

number_column_params = {
    "data_type": "number",
    "column_type": "source",
    **common_params,
}


@pytest.fixture
def sample_columns():
    return [
        ColumnNode(
            id="test::col1",
            name="col1",
            **base_column_params,
        ),
        ColumnNode(
            id="test::col2",
            name="col2",
            data_type="number",
            column_type="derived",
            **common_params,
            expression=pl.col("col1").cast(pl.Int64).alias("col2"),
        ),
    ]


@pytest.fixture
def base_column():
    return ColumnNode(
        id="test::base",
        name="base",
        **base_column_params,
    )


def test_registry_initialization():
    # Test empty initialization
    registry = ColumnRegistry()
    assert len(registry.by_id) == 0
    assert len(registry.by_name) == 0


def test_registry_with_columns(sample_columns):
    # Test initialization with columns
    registry = ColumnRegistry(sample_columns)
    assert len(registry.by_id) == 2
    assert len(registry.by_name) == 2

    # Test column retrieval - comparing relevant attributes instead of direct equality
    col1 = registry.get_by_id("test::col1")
    assert col1.id == sample_columns[0].id
    assert col1.name == sample_columns[0].name
    assert col1.data_type == sample_columns[0].data_type

    col2 = registry.get_by_name("col2")
    assert col2.id == sample_columns[1].id
    assert col2.name == sample_columns[1].name
    assert col2.data_type == sample_columns[1].data_type


def test_registry_append_and_extend(sample_columns):
    registry = ColumnRegistry()

    # Test append
    registry.append(sample_columns[0])
    assert len(registry.by_id) == 1
    col1 = registry.get_by_name("col1")
    assert col1.id == sample_columns[0].id
    assert col1.name == sample_columns[0].name

    # Test extend
    registry.extend([sample_columns[1]])
    assert len(registry.by_id) == 2
    col2 = registry.get_by_id("test::col2")
    assert col2.id == sample_columns[1].id
    assert col2.name == sample_columns[1].name


def test_registry_remove(sample_columns):
    registry = ColumnRegistry(sample_columns)

    # Test remove
    registry.remove(sample_columns[0])
    assert len(registry.by_id) == 1
    assert registry.get_by_id("test::col1") is None
    assert registry.get_by_name("col1") is None


def test_registry_map_parent_ids(sample_columns):
    registry = ColumnRegistry(sample_columns)

    # Test mapping source column names to IDs
    col2 = registry.get_by_id("test::col2")

    # Check that parent_columns was auto-populated from expression
    assert "col1" in col2.parent_columns
    # Check that parent_ids was populated by registry from parent_columns
    assert "test::col1" in col2.parent_ids


def test_registry_catch_duplicate_ids(sample_columns):
    registry = ColumnRegistry(sample_columns)

    # Test duplicate ID
    with pytest.raises(
        AssertionError, match=f"Duplicate column ID: {sample_columns[0].id}"
    ):
        registry.extend([sample_columns[0]])  # Adding duplicate


def test_registry_get_execution_stages(sample_columns):
    registry = ColumnRegistry(sample_columns)

    # Test execution stages
    stages = registry.get_execution_stages()
    assert len(stages) == 1  # Should have 1 stage due to dependency
    assert stages[0][0].id == "test::col2"  # First stage should have col2


def test_column_node_expression_name_mismatch():
    # Test that expression output name must match column name
    with pytest.raises(
        AssertionError,
        match="Column name wrong_name does not match expression output name col2",
    ):
        ColumnNode(
            id="test::col2",
            name="wrong_name",  # Doesn't match expression output
            **number_column_params,
            expression=pl.col("col1").cast(pl.Int64).alias("col2"),
        )


def test_column_node_parent_columns_auto_population():
    # Test that parent_columns are auto-populated from expression
    col = ColumnNode(
        id="test::col2",
        name="col2",
        **number_column_params,
        expression=pl.col("col1").cast(pl.Int64).alias("col2"),
    )
    assert col.parent_columns == frozenset({"col1"})

    # Test that explicit parent_columns are respected
    col_explicit = ColumnNode(
        id="test::col2",
        name="col2",
        **number_column_params,
        expression=pl.col("col1").cast(pl.Int64).alias("col2"),
        parent_columns=frozenset({"explicit_source"}),
    )
    assert col_explicit.parent_columns == frozenset({"explicit_source"})


def test_column_node_with_parent_ids():
    # Test that with_parent_ids creates a new instance with updated parent_ids
    col = ColumnNode(
        id="test::col2",
        name="col2",
        **number_column_params,
        expression=pl.col("col1").cast(pl.Int64).alias("col2"),
    )

    new_col = col.with_parent_ids({"test::col1"})
    assert new_col.parent_ids == frozenset({"test::col1"})
    assert col.parent_ids == frozenset()  # Original should be unchanged


def test_registry_with_missing_parent_columns():
    # Test registry behavior when a column references a non-existent source
    col_with_mixed_sources = ColumnNode(
        id="test::col2",
        name="col2",
        **number_column_params,
        parent_columns=frozenset({"existing_col", "non_existent_col"}),
    )

    existing_col = ColumnNode(
        id="test::existing_col",
        name="existing_col",
        **base_column_params,
    )

    registry = ColumnRegistry([existing_col, col_with_mixed_sources])

    # Source columns should retain both sources
    assert col_with_mixed_sources.parent_columns == frozenset(
        {"existing_col", "non_existent_col"}
    )

    # But parent_ids should only contain the ID of the existing source
    col2 = registry.get_by_id("test::col2")
    assert col2.parent_ids == frozenset({"test::existing_col"})
    assert "test::non_existent_col" not in col2.parent_ids


def test_registry_integrity_unmapped_sources(base_column):
    # Column with unmapped source
    col_unmapped = ColumnNode(
        id="test::unmapped",
        name="unmapped",
        **number_column_params,
        parent_columns=frozenset({"non_existent"}),
    )

    registry = ColumnRegistry([base_column, col_unmapped])
    # Registry should set parent_ids to empty since source doesn't exist
    assert registry.get_by_id("test::unmapped").parent_ids == frozenset()
    with pytest.raises(ValueError, match="Unmapped source columns found"):
        registry.verify_integrity()


def test_registry_integrity_invalid_parent_ids(base_column):
    # Column with source that exists but incorrect ID mapping
    col_with_source = ColumnNode(
        id="test::derived",
        name="derived",
        data_type="number",
        column_type="derived",
        **common_params,
        parent_columns=frozenset({"base"}),
    )

    registry = ColumnRegistry([base_column, col_with_source])
    # Registry should set the correct source_id
    assert registry.get_by_id("test::derived").parent_ids == frozenset({"test::base"})
    assert registry.verify_integrity() is True


def test_registry_integrity_incorrect_mappings(base_column):
    # Column referencing base but with no parent_columns set
    col_incorrect = ColumnNode(
        id="test::incorrect",
        name="incorrect",
        **number_column_params,
        expression=pl.col("base").alias(
            "incorrect"
        ),  # References base but parent_columns will be auto-populated
    )

    registry = ColumnRegistry([base_column, col_incorrect])
    # Check that parent_columns was auto-populated from expression
    assert registry.get_by_id("test::incorrect").parent_columns == frozenset({"base"})
    # Check that parent_ids was properly mapped
    assert registry.get_by_id("test::incorrect").parent_ids == frozenset({"test::base"})
    assert registry.verify_integrity() is True


def test_registry_integrity_multiple_violations(base_column):
    # Column with mix of valid and invalid sources
    col_mixed = ColumnNode(
        id="test::mixed",
        name="mixed",
        **number_column_params,
        parent_columns=frozenset({"base", "non_existent1", "non_existent2"}),
    )

    registry = ColumnRegistry([base_column, col_mixed])
    # Registry should only include the valid source ID
    assert registry.get_by_id("test::mixed").parent_ids == frozenset({"test::base"})
    # But verify should fail due to unmapped sources
    with pytest.raises(ValueError) as exc:
        registry.verify_integrity()

    error_msg = str(exc.value)
    assert "Unmapped source columns found" in error_msg
    assert "non_existent1" in error_msg
    assert "non_existent2" in error_msg


def test_registry_integrity_valid_case(base_column):
    # Test valid case with proper source mapping
    valid_col = ColumnNode(
        id="test::valid",
        name="valid",
        **number_column_params,
        parent_columns=frozenset({"base"}),
    )

    registry = ColumnRegistry([base_column, valid_col])
    # Check that parent_ids was properly set by registry
    assert registry.get_by_id("test::valid").parent_ids == frozenset({"test::base"})
    assert registry.verify_integrity() is True


def test_column_node_expression_validation():
    # Valid case - expression output name matches column name
    valid_col = ColumnNode(
        id="test::col1",
        name="col1",
        **number_column_params,
        expression=pl.col("base").alias("col1"),
    )
    assert valid_col.parent_columns == frozenset({"base"})

    # Valid case - explicit parent_columns are respected even with expression
    valid_explicit = ColumnNode(
        id="test::col1",
        name="col1",
        **number_column_params,
        expression=pl.col("base").alias("col1"),
        parent_columns=frozenset({"explicit_source"}),
    )
    assert valid_explicit.parent_columns == frozenset({"explicit_source"})

    # Invalid case - expression output name doesn't match column name
    with pytest.raises(
        AssertionError,
        match="Column name wrong_name does not match expression output name col1",
    ):
        ColumnNode(
            id="test::wrong",
            name="wrong_name",
            **number_column_params,
            expression=pl.col("base").alias("col1"),
        )

    # Invalid case - expression with no alias
    with pytest.raises(
        AssertionError,
        match="Column name col1 does not match expression output name base",
    ):
        ColumnNode(
            id="test::col1",
            name="col1",
            **number_column_params,
            expression=pl.col("base"),  # No alias set
        )

    # Test chained expression
    chained_col = ColumnNode(
        id="test::col1",
        name="col1",
        data_type="number",
        sensitivity=Sensitivity.NON_SENSITIVE,
        schema="test",
        stage="test_stage",
        expression=pl.col("base").cast(pl.Int64).alias("col1"),
        column_type="derived",
    )
    assert chained_col.parent_columns == frozenset({"base"})

    # Test multiple source columns
    multi_source = ColumnNode(
        id="test::sum",
        name="sum",
        data_type="number",
        sensitivity=Sensitivity.NON_SENSITIVE,
        schema="test",
        stage="test_stage",
        expression=(pl.col("a") + pl.col("b")).alias("sum"),
        column_type="derived",
    )
    assert multi_source.parent_columns == frozenset({"a", "b"})



import polars as pl
from datetime import date


def calculate_age(
    col: str,
    *,
    comparison_date: date | str | pl.Expr = date.today(),
    alias: str = "age",
):
    if isinstance(comparison_date, date):
        return (
            pl.lit(comparison_date.year)
            - pl.col(col).dt.year()
            - (
                (pl.lit(comparison_date.month) < pl.col(col).dt.month())
                | (
                    (pl.lit(comparison_date.month) == pl.col(col).dt.month())
                    & (pl.lit(comparison_date.day) < pl.col(col).dt.day())
                )
            )
        ).alias(alias)

    elif isinstance(comparison_date, str):
        return (
            pl.col(comparison_date).dt.year()
            - pl.col(col).dt.year()
            - (
                (pl.col(comparison_date).dt.month() < pl.col(col).dt.month())
                | (
                    (pl.col(comparison_date).dt.month() == pl.col(col).dt.month())
                    & (pl.col(comparison_date).dt.day() < pl.col(col).dt.day())
                )
            )
        ).alias(alias)

    elif isinstance(comparison_date, pl.Expr):
        return (
            comparison_date.dt.year()
            - pl.col(col).dt.year()
            - (
                (comparison_date.dt.month() < pl.col(col).dt.month())
                | (
                    (comparison_date.dt.month() == pl.col(col).dt.month())
                    & (comparison_date.dt.day() < pl.col(col).dt.day())
                )
            )
        ).alias(alias)

    else:
        raise ValueError(f"Invalid comparison_date type: {type(comparison_date)}")



import pytest
from zeolite.ref import parse_column_name, ColumnRef
from zeolite._utils.column_id import COLUMN_NAME_PART_DIVIDER as PD, COLUMN_NAME_SECTION_DIVIDER as SD

def test_simple_base_name():
    """Test parsing a simple base name"""
    result = parse_column_name("foo")
    assert result == ColumnRef(
        base_name="foo",
        is_clean=False,
        is_derived=False,
        is_meta=False,
        is_custom_check=False,
        check_name=None
    )

def test_base_name_with_sections():
    """Test parsing a base name with multiple sections"""
    result = parse_column_name(f"foo{SD}bar{SD}baz")
    assert result == ColumnRef(
        base_name=f"foo{SD}bar{SD}baz",
        is_clean=False,
        is_derived=False,
        is_meta=False,
        is_custom_check=False,
        check_name=None
    )

def test_valid_single_prefix():
    """Test parsing with a single valid prefix"""
    result = parse_column_name(f"meta{SD}foo")
    assert result == ColumnRef(
        base_name="foo",
        is_clean=False,
        is_derived=False,
        is_meta=True,
        is_custom_check=False,
        check_name=None
    )

def test_valid_prefixes():
    """Test parsing with valid prefixes"""
    result = parse_column_name(f"meta{PD}derived{SD}foo")
    assert result == ColumnRef(
        base_name="foo",
        is_clean=False,
        is_derived=True,
        is_meta=True,
        is_custom_check=False,
        check_name=None
    )

def test_invalid_prefix_section():
    """Test parsing when first section contains non-prefix parts"""
    result = parse_column_name(f"foo{SD}bam")
    assert result == ColumnRef(
        base_name=f"foo{SD}bam",
        is_clean=False,
        is_derived=False,
        is_meta=False,
        is_custom_check=False,
        check_name=None
    )

def test_valid_suffixes():
    """Test parsing with valid suffixes"""
    result = parse_column_name(f"foo{SD}clean{PD}check{PD}is_empty")
    assert result == ColumnRef(
        base_name="foo",
        is_clean=True,
        is_derived=False,
        is_meta=False,
        is_custom_check=False,
        check_name="is_empty"
    )

def test_invalid_suffix_section():
    """Test parsing when last section contains non-suffix parts"""
    test_cases = [
        f"foo{SD}clean{PD}popsicle",
        f"foo{SD}clean{PD}popsicle{PD}check{PD}is_empty",
        f"foo{SD}clean{PD}check{PD}is_empty{PD}popsicle"
    ]
    
    for test_case in test_cases:
        result = parse_column_name(test_case)
        assert result == ColumnRef(
            base_name=test_case,
            is_clean=False,
            is_derived=False,
            is_meta=False,
            is_custom_check=False,
            check_name=None
        )

def test_complex_cases():
    """Test parsing complex combinations"""
    test_cases = [
        (
            f"meta{PD}derived{SD}foo{SD}bar{SD}clean{PD}check{PD}is_empty",
            ColumnRef(
                base_name=f"foo{SD}bar",
                is_clean=True,
                is_derived=True,
                is_meta=True,
                is_custom_check=False,
                check_name="is_empty"
            )
        ),
        (
            f"meta{SD}bam{SD}clean{PD}check{PD}is_empty",
            ColumnRef(
                base_name="bam",
                is_clean=True,
                is_derived=False,
                is_meta=True,
                is_custom_check=False,
                check_name="is_empty"
            )
        ),
        (
            f"bam{SD}popsicle",
            ColumnRef(
                base_name=f"bam{SD}popsicle",
                is_clean=False,
                is_derived=False,
                is_meta=False,
                is_custom_check=False,
                check_name=None
            )
        )
    ]
    
    for test_case, expected in test_cases:
        result = parse_column_name(test_case)
        assert result == expected, f"Failed for {test_case}"

def test_multiple_valid_prefixes():
    """Test parsing with multiple valid prefixes in first section"""
    test_cases = [
        (
            f"meta{PD}derived{PD}custom_check{SD}foo",
            ColumnRef(
                base_name="foo",
                is_clean=False,
                is_derived=True,
                is_meta=True,
                is_custom_check=True,
                check_name=None
            )
        ),
        (
            f"derived{PD}meta{SD}foo",
            ColumnRef(
                base_name="foo",
                is_clean=False,
                is_derived=True,
                is_meta=True,
                is_custom_check=False,
                check_name=None
            )
        )
    ]
    
    for test_case, expected in test_cases:
        result = parse_column_name(test_case)
        assert result == expected, f"Failed for {test_case}"

def test_multiple_valid_suffixes():
    """Test parsing with multiple valid suffixes in last section"""
    test_cases = [
        (
            f"foo{SD}clean{PD}check{PD}is_empty{PD}check{PD}is_null",
            ColumnRef(
                base_name="foo",
                is_clean=True,
                is_derived=False,
                is_meta=False,
                is_custom_check=False,
                check_name="is_null"  # Last check takes precedence
            )
        ),
        (
            f"foo{SD}check{PD}is_empty{PD}clean",
            ColumnRef(
                base_name="foo",
                is_clean=True,
                is_derived=False,
                is_meta=False,
                is_custom_check=False,
                check_name="is_empty"
            )
        )
    ]
    
    for test_case, expected in test_cases:
        result = parse_column_name(test_case)
        assert result == expected, f"Failed for {test_case}"

def test_mixed_valid_invalid_parts():
    """Test parsing with mixed valid and invalid parts"""
    test_cases = [
        (
            f"meta{PD}invalid{SD}foo{SD}clean{PD}invalid{PD}check{PD}is_empty",
            ColumnRef(
                base_name=f"meta{PD}invalid{SD}foo{SD}clean{PD}invalid{PD}check{PD}is_empty",
                is_clean=False,
                is_derived=False,
                is_meta=False,
                is_custom_check=False,
                check_name=None
            )
        ),
        (
            f"invalid{PD}meta{SD}foo{SD}check{PD}invalid{PD}is_empty",
            ColumnRef(
                base_name=f"invalid{PD}meta{SD}foo{SD}check{PD}invalid{PD}is_empty",
                is_clean=False,
                is_derived=False,
                is_meta=False,
                is_custom_check=False,
                check_name=None
            )
        )
    ]
    
    for test_case, expected in test_cases:
        result = parse_column_name(test_case)
        assert result == expected, f"Failed for {test_case}"

def test_empty_parts():
    """Test parsing with empty parts between dividers"""
    test_cases = [
        (
            f"meta{SD}{SD}foo",  # Double section divider
            ColumnRef(
                base_name="foo",
                is_clean=False,
                is_derived=False,
                is_meta=True,
                is_custom_check=False,
                check_name=None
            )
        ),
        (
            f"foo{SD}{SD}clean",  # Double section divider
            ColumnRef(
                base_name="foo",
                is_clean=True,
                is_derived=False,
                is_meta=False,
                is_custom_check=False,
                check_name=None
            )
        )
    ]
    
    for test_case, expected in test_cases:
        result = parse_column_name(test_case)
        assert result == expected, f"Failed for {test_case}"

def test_special_characters():
    """Test parsing with special characters in names"""
    test_cases = [
        (
            f"foo.bar{SD}clean",
            ColumnRef(
                base_name="foo.bar",
                is_clean=True,
                is_derived=False,
                is_meta=False,
                is_custom_check=False,
                check_name=None
            )
        ),
        (
            f"foo-bar{SD}check{PD}is-empty",
            ColumnRef(
                base_name="foo-bar",
                is_clean=False,
                is_derived=False,
                is_meta=False,
                is_custom_check=False,
                check_name="is-empty"
            )
        )
    ]
    
    for test_case, expected in test_cases:
        result = parse_column_name(test_case)
        assert result == expected, f"Failed for {test_case}" 


from dataclasses import dataclass, asdict, field
from typing import Optional, TYPE_CHECKING, Union

from polars import Expr, col

from .._utils.column_id import (
    make_column_id,
    ID_PART_DIVIDER,
    COLUMN_NAME_PART_DIVIDER,
    COLUMN_NAME_SECTION_DIVIDER,
    ID_SECTION_DIVIDER,
    ColumnIdConfig, parse_column_name_to_config,
)

if TYPE_CHECKING:
    from ..registry import ColumnRegistry
    from ..types import ColumnNode
    from ..column.validation import _BaseRowCheck, ColumnCheckType


@dataclass(frozen=True)
class ColumnRef(ColumnIdConfig):
    """Reference to a column that can be used to look up actual column names/ids"""
    # alias_name: Optional[str] = None

    def _chain(self, **kwargs: dict[str, any]) -> "ColumnRef":
        """Create a new instance with some fields replaced"""
        return ColumnRef(**{**asdict(self), **kwargs})

    def clean(self) -> "ColumnRef":
        return self._chain(is_clean=True)

    def check(
        self, validation: Union["ColumnCheckType", type["_BaseRowCheck"], str]
    ) -> "ColumnRef":
        if isinstance(validation, str):
            check_name = validation
        elif isinstance(validation, type):
            check_name = validation.method_id()
        else:
            check_name = validation.method_id()
        return self._chain(is_check=True, check_name=check_name)

    def derived(self) -> "ColumnRef":
        return self._chain(is_derived=True)

    def custom_check(self) -> "ColumnRef":
        return self._chain(is_custom_check=True, is_derived=True)

    def with_schema(self, schema: str, stage: str | None = None) -> "ColumnRef":
        if stage is not None:
            return self._chain(schema=schema, stage=stage)
        else:
            return self._chain(schema=schema)

    def with_base_name(self, name: str) -> "ColumnRef":
        """Sets the column reference name"""
        ref = parse_column_name(name)
        return self._chain(base_name=ref.base_name)

    # def alias(self, name: str) -> 'ColumnRef':
    #     """Set an alias for this column reference"""
    #     return self._chain(alias_name=name)

    def get_id(
        self,
        *,
        schema: str | None = None,
        stage: str | None = None,
        include_schema_prefix: bool = True,
    ) -> str:
        """Get the ID of the column reference"""
        return make_column_id(
            name=self.base_name,
            schema=schema if schema else self.schema,
            stage=stage if stage else self.stage,
            is_clean=self.is_clean,
            is_derived=self.is_derived,
            is_meta=self.is_meta,
            is_custom_check=self.is_custom_check,
            check_name=self.check_name,
            include_schema_prefix=include_schema_prefix,
        )

    def get_name(self) -> str:
        """Get the name of the column reference"""
        return (
            self.get_id(include_schema_prefix=False)
            .replace(ID_SECTION_DIVIDER, COLUMN_NAME_SECTION_DIVIDER)
            .replace(ID_PART_DIVIDER, COLUMN_NAME_PART_DIVIDER)
        )

    @property
    def col(self) -> Expr:
        """Get the name of the column reference"""
        return col(self.name)

    @property
    def name(self) -> str:
        """Get the name of the column reference"""
        return self.get_name()

    @property
    def id(self) -> str:
        """Get the ID of the column reference"""
        return self.get_id()

    @classmethod
    def from_id(cls, id: str) -> "ColumnRef":
        """Create a new column reference from an ID"""
        # Split parts
        parts = id.split(ID_PART_DIVIDER)

        # TODO: This is a mess & doesn't work for "meta" columns yet - need to align/replace parse_column_id
        # Initialize variables
        stage = None
        schema = None
        name = None
        is_clean = False
        is_derived = False
        is_custom_check = False
        check_name = None

        # Parse based on number of parts
        if len(parts) == 1:
            # Simple case: just name
            name = parts[0]

        else:
            # Check if starts with stage
            current_idx = 0
            if len(parts) >= 3 and not parts[0].startswith("derived"):
                stage = parts[0]
                current_idx += 1

            # Check for schema
            if len(parts) >= (2 + current_idx):
                schema = parts[current_idx]
                current_idx += 1

            # Check for derived
            if "derived" in parts[current_idx:]:
                is_derived = True
                derived_idx = parts.index("derived", current_idx)
                current_idx = derived_idx + 1

                # Check for custom check
                if current_idx < len(parts) and parts[current_idx] == "custom_check":
                    check_name = parts[current_idx + 1]
                    return cls(
                        base_name=check_name,
                        schema=schema,
                        stage=stage,
                        is_derived=True,
                        is_custom_check=True,
                    )

            # Get remaining parts
            remaining = parts[current_idx:]

            # Parse remaining parts
            if "clean" in remaining:
                is_clean = True
                clean_idx = remaining.index("clean")
                name = remaining[clean_idx - 1] if clean_idx > 0 else remaining[-1]

                if "check" in remaining[clean_idx:]:
                    check_idx = remaining.index("check", clean_idx)
                    check_name = remaining[check_idx + 1]

            elif "check" in remaining:
                check_idx = remaining.index("check")
                name = remaining[check_idx - 1] if check_idx > 0 else remaining[-1]
                check_name = remaining[check_idx + 1]

            else:
                name = remaining[-1]

        return cls(
            base_name=name,
            schema=schema,
            stage=stage,
            is_clean=is_clean,
            is_derived=is_derived,
            check_name=check_name,
        )

    def resolve(
        self,
        registry: "ColumnRegistry",
        *,
        schema: str | None = None,
        stage: str | None = None,
    ) -> "ColumnNode":
        """Resolve the reference in the registry to an actual column node"""
        col_id = self.get_id(schema=schema, stage=stage)

        node = registry.get_by_id(col_id)
        if not node:
            raise ValueError(f"Could not find column with ID {col_id} in registry")
        return node


@dataclass(frozen=True)
class MetaColumnRef(ColumnRef):
    """Reference to a derived column"""

    is_meta: bool = field(default=True, init=False)


@dataclass(frozen=True)
class DerivedColumnRef(ColumnRef):
    """Reference to a derived column"""

    is_derived: bool = field(default=True, init=False)


@dataclass(frozen=True)
class CustomCheckColumnRef(ColumnRef):
    """Reference to a custom check column"""

    is_custom_check: bool = field(default=True, init=False)
    is_derived: bool = field(default=True, init=False)


def get_column_ref_from_id(col_id: str) -> ColumnRef:
    """Get a column reference from an ID"""
    return ColumnRef.from_id(col_id)


def parse_column_name(col_name: str) -> ColumnRef:
    """
        Parse a column name string into a ColumnRef

        Args:
            col_name: The column name string to parse

        Returns:
            ColumnRef
        """
    config = parse_column_name_to_config(col_name)
    return ColumnRef(**asdict(config))
        



# %%
from dataclasses import dataclass, field, KW_ONLY, fields, replace
from typing import Any, List, Literal
from polars import Expr, Float64, Int64

from ...ref import ColumnRef
from ...types import ColumnNode
from ...types.sensitivity import Sensitivity
from ...types.data_type import ColumnDataType
from ...types.validation.threshold import CheckLevel, ThresholdLevel
from ..._utils.args import flatten_args
from .._clean import (
    CleanStage,
    CleanBooleanColumn,
    CleanDateColumn,
    CleanEnumColumn,
    CleanIdColumn,
    CleanNumberColumn,
    StringColumn,
    CleanColumn,
)
from ..validation import ColumnCheckType
from ..validation.check_base import RowCheckBase, ThresholdType
from ..validation.validation_node import create_validation_rule

# %%

type _RequiredLevel = CheckLevel | bool | None


@dataclass(frozen=True)
class _ColumnParams:
    ref: ColumnRef
    _: KW_ONLY
    data_type: ColumnDataType = field(default="unknown")
    sensitivity: Sensitivity = field(default=Sensitivity.UNKNOWN)
    aliases: set[str] = field(default_factory=set)
    if_missing: ThresholdLevel = field(default=ThresholdLevel.PASS)
    validations: List[ColumnCheckType] = field(default_factory=list)
    clean: CleanStage | None = None
    expression: Expr | None = None
    custom_validation: dict[str, str | ThresholdType] | None = None


def _verify_validations(validations: List[ColumnCheckType] | None):
    if validations is not None:
        for v in validations:
            assert isinstance(v, RowCheckBase), (
                f"Expected {v} to be a Column Validation Check"
            )


class ColumnSchema:
    """
    Defines a column in a table schema, including data type, sensitivity, validations, and cleaning steps.

    Parameters:
        col_ref (ColumnRef): Reference to the column.
        data_type (ColumnDataType): Data type of the column.
        sensitivity (Sensitivity): Sensitivity level.
        aliases (set[str]): Aliases for the column.
        validations (List[ColumnCheckType]): Validation checks.
        clean (CleanStage | ...): Cleaning stage or type.
    """

    def __init__(
        self,
        col_ref: ColumnRef,
        *,
        data_type: ColumnDataType = "unknown",
        sensitivity: Sensitivity = None,
        aliases: set[str] = None,
        required: _RequiredLevel = False,
        validations: List[ColumnCheckType] = None,
        clean: CleanStage | ColumnDataType | dict[str, Any] | None = None,
    ):
        _verify_validations(validations)
        if clean is not None:
            if isinstance(clean, CleanColumn):
                pass
            elif isinstance(clean, str):
                clean = self._get_clean_stage(clean)
            elif isinstance(clean, dict):
                clean = self._get_clean_stage(**clean)
            else:
                clean = None

        self._params = _ColumnParams(
            ref=col_ref,
            data_type=data_type,
            sensitivity=sensitivity if sensitivity is not None else Sensitivity.UNKNOWN,
            aliases=aliases if aliases is not None else set(),
            validations=validations if validations is not None else [],
            clean=clean,
            if_missing=_get_required_level(required),
        )

    def sensitivity(self, sensitivity: Sensitivity) -> "ColumnSchema":
        return self._replace(sensitivity=sensitivity)

    def data_type(self, data_type: ColumnDataType) -> "ColumnSchema":
        return self._replace(data_type=data_type)

    def aliases(self, *args: str | list[str], merge: bool = False) -> "ColumnSchema":
        aliases = set(flatten_args(args))
        if merge:
            aliases = aliases | self._params.aliases
        return self._replace(aliases=aliases)

    def validations(
        self, *args: ColumnCheckType | list[ColumnCheckType], merge: bool = False
    ) -> "ColumnSchema":
        validations = flatten_args(args)
        _verify_validations(validations)
        if merge:
            validations = validations + self._params.validations

        return self._replace(validations=validations)

    def clean(self, to: ColumnDataType | CleanStage, **kwargs: Any) -> "ColumnSchema":
        return self._replace(clean=self._get_clean_stage(to, **kwargs))

    def required(self, required: _RequiredLevel = True) -> "ColumnSchema":
        return self._replace(if_missing=_get_required_level(required))

    def get_nodes(self, schema: str, stage: str | None = None) -> List[ColumnNode]:
        p = self._params
        ref = p.ref.with_schema(schema, stage)

        column_type = _get_col_type(ref)
        validation_rule = (
            create_validation_rule(
                check_method_id="custom_validation",
                message=p.custom_validation["message"],
                source_column=", ".join(p.expression.meta.root_names()),
                alias=ref.name,
                schema=ref.schema,
                thresholds=p.custom_validation["thresholds"],
            )
            if ref.is_custom_check
            else None
        )

        nodes: list[ColumnNode] = [
            ColumnNode(
                id=ref.id,
                name=ref.name,
                data_type=p.data_type,
                column_type=column_type,
                schema=ref.schema,
                stage=ref.stage,
                sensitivity=p.sensitivity,
                expression=self.expr,
                validation_rule=validation_rule,
            )
        ]
        if p.clean is not None:
            clean_expr = p.clean.clean_expr(ref)
            nodes.append(
                ColumnNode(
                    id=ref.clean().id,
                    name=clean_expr.meta.output_name(),
                    data_type=p.clean.data_type,
                    column_type="cleaned",
                    schema=ref.schema,
                    stage=ref.stage,
                    sensitivity=p.clean.col_sensitivity
                    if p.clean.col_sensitivity is not None
                    else p.sensitivity,
                    expression=clean_expr,
                    validation_rule=None,
                )
            )
        for v in p.validations:
            nodes.append(v.get_validation_node(ref))

        return nodes

    @property
    def ref(self) -> ColumnRef:
        """Get the name of the column reference"""
        return self._params.ref

    @property
    def get_aliases(self) -> set[str]:
        """Get the aliases of the column reference"""
        return self._params.aliases

    @property
    def if_missing(self) -> ThresholdLevel:
        """Get the threshold level for the column if missing"""
        return self._params.if_missing

    @property
    def has_expression(self) -> bool:
        """Check if the column is/has an expression"""
        return self._params.expression is not None

    @property
    def expr(self) -> Expr | None:
        return self._params.expression.alias(self._params.ref.name) if self.has_expression else None

    def with_name(self, name: str):
        return self._replace(ref=self._params.ref.with_base_name(name))

    def derived(self, expr: Expr) -> "ColumnSchema":
        return self._replace(ref=self._params.ref.derived(), expression=expr)

    def custom_check(
        self, expr: Expr, thresholds: ThresholdType, message: str
    ) -> "ColumnSchema":
        return self._replace(
            ref=self._params.ref.custom_check(),
            expression=expr,
            custom_validation={"message": message, "thresholds": thresholds},
        )

    def _get_clean_stage(
        self, to: ColumnDataType | CleanStage, **kwargs: dict[str, Any]
    ) -> CleanStage:
        # TODO: figure out a cleaner (ha) way to do determine clean stage based on data type
        # FIXME: figure out if it's possible to have dynamic typed kwargs based on the data type

        if to is None:
            raise ValueError("'to' cannot be None for cleaning stage")

        if isinstance(to, CleanColumn):
            return to
        elif to == "string":
            return StringColumn(**kwargs)
        elif to == "sanitised_string":
            return StringColumn(
                **kwargs,
                sanitize=kwargs["sanitize"] if "sanitize" in kwargs else "full",
            )
        elif to == "number":
            return CleanNumberColumn(**kwargs)
        elif to == "integer":
            return CleanNumberColumn(output_format=Int64, data_type="integer", **kwargs)
        elif to == "decimal" or to == "float":
            return CleanNumberColumn(
                output_format=Float64,
                data_type="float",
                **kwargs,
            )
        elif to == "boolean":
            return CleanBooleanColumn(**kwargs)
        elif to == "date":
            return CleanDateColumn(**kwargs)
        # elif data_type == "datetime":
        #     return replace(self, clean=CleanDateTimeColumn(**kwargs))
        elif to == "enum":
            return CleanEnumColumn(**kwargs)
        elif to == "id":
            return CleanIdColumn(**kwargs)

        # TODO: implement custom clean stage
        # elif to == "custom":
        # return CleanCustomColumn(**kwargs)
        else:
            raise ValueError(f"Unimplemented clean stage: {to}")

    def _replace(self, **kwargs):
        new_params = replace(self._params, **kwargs)
        # initiate an empty column reference & manually apply the new params
        # we do it this way to set properties not allowed in the constructor
        return ColumnSchema(col_ref=new_params.ref).__set_params(new_params)

    def __set_params(self, params: _ColumnParams) -> "ColumnSchema":
        self._params = params
        return self

    def __repr__(self):
        """Return a string representation of the instance."""
        param_strs = []
        for field_name in [f.name for f in fields(_ColumnParams)]:
            value = getattr(self._params, field_name)
            if value is not None and (not isinstance(value, list) or value):
                param_strs.append(f"{field_name}={repr(value)}")
        return f"{self.__class__.__name__}({', '.join(param_strs)})"


# %%
def _get_col_type(
    ref: ColumnRef,
) -> Literal["source", "cleaned", "validation", "meta", "derived", "custom_validation"]:
    if ref.is_meta:
        return "meta"
    if ref.is_derived:
        return "custom_validation"
    elif ref.is_derived:
        return "derived"
    else:
        return "source"


def _get_required_level(required: _RequiredLevel) -> ThresholdLevel:
    if isinstance(required, bool):
        return ThresholdLevel.REJECT if required else ThresholdLevel.PASS
    elif required == "debug":
        return ThresholdLevel.DEBUG
    elif required == "warning":
        return ThresholdLevel.WARNING
    elif required == "error":
        return ThresholdLevel.ERROR
    elif required == "reject":
        return ThresholdLevel.REJECT
    else:
        return ThresholdLevel.PASS


# %%



from dataclasses import KW_ONLY, dataclass
import polars as pl

from .threshold import Threshold, ThresholdLevel
from ..error import DataValidationError

type CheckValue = (
    str
    | int
    | float
    | bool
    | None
    | list[str]
    | list[int]
    | list[float]
    | list[bool]
    | set[str]
    | set[int]
    | set[float]
    | set[bool]
)


@dataclass(frozen=True)
class ColumnValidationRule:
    check_id: str
    _: KW_ONLY
    thresholds: Threshold
    message: str
    source_column: str
    check_column: str
    schema: str

    def validate(
        self, lf: pl.LazyFrame, source: str | None = None
    ) -> DataValidationError | None:
        checks = lf.select(
            pl.len().alias("total_rows"),
            pl.col(self.check_column)
            .ne(ThresholdLevel.PASS.level)
            .sum()
            .alias("failed_rows"),
        ).collect()

        total_rows = checks["total_rows"].item()
        failed_rows = checks["failed_rows"].item()

        if failed_rows == 0:
            return None

        res = self.thresholds.resolve(failed_rows=failed_rows, total_rows=total_rows)

        return (
            DataValidationError(
                schema=self.schema,
                source=source,
                column=self.check_column,
                error=self.check_id,
                level=res.level,
                fraction_failed=res.fraction_failed,
                count_failed=res.count_failed,
                message=self._format_message(
                    self.source_column, failed_rows, total_rows
                ),
            )
            if res.level != "pass"
            else None
        )

    def _format_message(
        self,
        col: str,
        count: int,
        total: int,
        value: str | int | float | bool | None = None,
    ) -> str:
        return (
            self.message.replace("{{column}}", f"`{col}`")
            .replace("{{count}}", f"{count:,}")
            .replace("{{fraction}}", f"{count / total:,.2%}")
        )
